# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 17:47:28 2020

@author: willi
"""

import warnings
import numpy as np


class ProbeVectorFactory():
    def __init__(self):
        pass

    def get_probe_vec(self, tag_epitope_dict, length):
        '''
        Generate a probe vector given a tag epitope dictionary and the Length
        of the transcript:

        Examples
        --------

        >>> probef.get_probe_vec({Tag_1:[10,15]},20)
        np.array([[0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,2,2,2,2,2]])


        Parameters
        ----------
        tag_epitope_dict : dictionary
            dictionary of tag epitopes, can be accessed from a POI.tag_epitopes.
        length : int
            Length of the transcript.

        Returns
        -------
        probe_vec : Ncolor x L Numpy array
            returns the cumulative sum probe vector per location of a ribosome,
            meaning: A ribosome at x location
            will have pv[x] intensity.

        '''
        pv = np.zeros((len(list(tag_epitope_dict)), length))
        for i in range(len(list(tag_epitope_dict))):
            pv[i, [tag_epitope_dict[list(tag_epitope_dict.keys())[i]]]] = 1
        pv = np.cumsum(pv, axis=1)
        return pv

    def get_probe_loc(self, tag_epitope_dict, length):
        '''
        Get 1, 0 probe location from a tag epitope dictionary

        Examples
        --------

        >>> probef.get_probe_loc({Tag_1:[10,15]},20)
        pv = np.array([[0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,1,0,0,0,0]])


        Parameters
        ----------
        tag_epitope_dict : dictionary
            dictionary of tag epitopes, can be accessed from a
            POI.tag_epitopes.
        length : int
            Length of the transcript.

        Returns
        -------
        probe_loc : Ncolor x L Numpy array
            returns the binary probe per location of an epitope,
        '''

        pv = np.zeros((len(list(tag_epitope_dict)), length))
        for i in range(len(list(tag_epitope_dict))):
            pv[i, [tag_epitope_dict[list(tag_epitope_dict.keys())[i]]]] = 1
        return pv

    @staticmethod
    def bin_probe_vecs(probe_loc, inds):
        '''
        given a probe location vector and an desired binning indices,
        return pv and pl binned

        Parameters
        ----------
        probe_loc : Ncolor by L binary numpy array
            locations of tag epitopes.
        inds : 1xL ind locations sizes
            binning strategy, the sum of this vector should = L:.

        Returns
        -------
        Probe_loc_binned: Ncolor x L numpy array
            the binned probe locations.
        Probe_vec_binned: Ncolor x L numpy array
            the binned probe intensity vector.
        '''
        probeloc_binned = np.zeros((probe_loc.shape[0], len(inds)-1))
        for i in range(0, len(inds)-1):
            probeloc_binned[:, i] = np.sum(probe_loc[:, inds[i]:inds[i+1]],
                                           axis=1)


        #probeloc_binned[:,-1] = np.sum(probe_loc[:,inds[-1]:],axis=1)
        probevec_binned = np.cumsum(probeloc_binned, axis=1)

        return probeloc_binned.astype(int), probevec_binned.astype(int)



    @staticmethod
    def intellegent_bin(pl, nbins, min_bin=1):
        '''
        Function to do intellegent binning, focuses resolution on the areas
        defined in the probe location vector

        Note if you pass it a minium bin that when min_bin*nbins
        > length of your sequence
        this function will fail

        Parameters
        ----------
        pl : numpy array
            numpy array of Ncolor x Length 0's 1's for probe locations.
        nbins : int
            number of desired bins.
        min_bin : int, optional
            DESCRIPTION. The default is 1.

        Returns
        -------
        pl_inds : numpy array
            n bin locations over the vector length given.

        '''

        if min_bin*nbins > pl.shape[1]:
            warnings.warn('Desired minimum binsize and desired number '\
                          'of bins is not possible with the length of the'\
                              ' probe vector, returning best guess')

        pl_inds = np.where(pl == 1)[1]
        if 0 not in pl_inds:
            pl_inds = np.hstack((np.array([0]), pl_inds))
        if pl.shape[1] not in pl_inds:
            pl_inds = np.hstack((pl_inds, np.array(pl.shape[1])))

        used_bins = len(pl_inds)
        k = len(pl_inds)-1
        j = 0
        to_add = []
        while used_bins < nbins+1:
            if j == k:

                j = 0

                prev_pl_inds = pl_inds
                pl_inds = np.hstack((pl_inds, np.array(to_add)))
                pl_inds = np.sort(pl_inds)

                if np.array_equal(prev_pl_inds, pl_inds):
                    break
                k = pl.shape[1]-1
                to_add = []

            newbin = int(pl_inds[j] + (pl_inds[j+1]-pl_inds[j])/2)

            if newbin not in pl_inds:
                if not (np.abs(pl_inds - newbin) <= min_bin).any():
                    to_add.append(newbin)
                    used_bins += 1
            j += 1

        pl_inds = np.hstack((pl_inds, np.array(to_add)))
        pl_inds = np.sort(pl_inds)

        return pl_inds

    @staticmethod
    def even_bin(length, nbins):
        '''
        Parameters
        ----------
        length : int
            Length of the vector to bin.
        nbins : int
            Number of desired bins.

        Returns
        -------
        inds : numpy array
            n bin locations over the vector length given.

        '''

        binsize = int(np.floor(length/nbins))

        k_lens = np.ones(nbins)*binsize

        to_redistribute = length%nbins

        k_lens[-to_redistribute:] = binsize+1

        inds = np.hstack(([0.], np.cumsum(k_lens))).astype(int)

        return inds



# -*- coding: utf-8 -*-
"""
Created on Wed Apr 29 16:01:24 2020

@author: willi
"""

import numpy as np
import ssa_translation_lowmem
import matplotlib.pyplot as plt
import time

# load the elongation 
kelong = np.loadtxt('elongationrates.txt')
kbind = kelong[0]
kcompl = kelong[-1]
kelong = kelong[1:-1]

ncolor = 2
t_array = np.array([0,10,20,30,50,100,250,500],dtype=np.float64)
t0 = 15
t_array = np.linspace(0,1000,1000,dtype=np.float64)
N_rib = 10
result = np.zeros((len(t_array)*ncolor),dtype=np.int32  )
#kelong = np.array([3.1,3.2,3.3,3.4,3.5,3.1,3.2,3.3,3.4,3.5],dtype=np.float64)
n_trajectories = 1
start = time.time()








#preallocated arrays here

all_ribosome_locations = np.zeros((n_trajectories,len(t_array),N_rib),dtype=np.int32)

all_results = np.zeros((n_trajectories,len(t_array),ncolor),dtype=np.int32)
lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t_array<20)[0]))

all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)

all_ribtimes = np.zeros((n_trajectories,400),dtype=np.float64)
all_coltimes = np.zeros((n_trajectories,400),dtype=np.int32)
nribs = np.array([0],dtype=np.int32)
all_ribs = np.zeros((n_trajectories,1))
seeds = np.random.randint(0,0x7FFFFFF,n_trajectories,dtype=np.int32)
x0 = np.zeros((N_rib),dtype=np.int32)

pl = np.zeros((2, len(kelong)), dtype=np.int32)

pl[:,[5,8,10,20,30,100,120,140]] = 1

pv = pl.astype(np.int32)
#pl[ [10,140],1  ] = 1

pl = np.cumsum(pl,axis=1)
pl = pl.astype(np.int32)
print(x0.shape)
print(all_results.shape)
print(all_frapresults.shape)
print(all_ribtimes.shape)
print(all_coltimes.shape)

all_col_points = []



kon = 3
koff = 3

k_probe = np.array([.8,.1],dtype=np.float64)

#record locations
flags = np.array([0,0,0],dtype=np.int32)
start = time.time()
for i in range(n_trajectories):
    result = np.zeros((ncolor,len(t_array)),dtype=np.int32)    
    ribloc = np.zeros((N_rib,len(t_array)),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    
    colpointsx = np.zeros(len(kelong)*400,dtype=np.int32)
    colpointst = np.zeros(len(kelong)*400,dtype=np.float64)

    print(result.dtype)
    print(ribtimes.dtype)
    print(coltimes.dtype)
    print(colpointsx.dtype)
    print(colpointst.dtype)
    print(kelong.dtype)
    print(frapresult.dtype)
    print(t_array.dtype)
    print(type(.03))
    print(type(kcompl))
    print(type(300))
    print(type(seeds[i]))
    print(nribs.dtype)
    print(x0.dtype)
    print(pl.dtype)
    print(type(kon))
    print(type(koff))
    print(k_probe.dtype)
    print(pv.dtype)
    print(flags.dtype)
    print(type(N_rib))

    ssa_translation_lowmem.run_SSA( result,ribtimes,coltimes,colpointsx,colpointst, kelong,frapresult,t_array,.03,kcompl, 0,0,0, seeds[i],nribs,x0,9, pl, 2, kon, koff,k_probe,pv,flags,N_rib)
    
    all_results[i,:,:] = result.T
    
    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    endcolrec = np.where(colpointsx == 0)[0][0]
    
    colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
    all_col_points.append(colpoints.T)

full_rec_t = time.time()-start
    
print('time for {0} trajectories {1}'.format(n_trajectories,time.time()-start))

plt.figure(dpi=300)
plt.plot(all_results[0],'bo')
plt.xlabel('time')
plt.ylabel('intensity')



start = time.time()
for i in range(n_trajectories):
    result = np.zeros((ncolor,len(t_array)),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    
    colpointsx = np.zeros(len(kelong)*400,dtype=np.int32)
    colpointst = np.zeros(len(kelong)*400,dtype=np.float64)
    
    

    ssa_translation_lowmem.run_SSA(result,ribtimes,coltimes,colpointsx,colpointst, kelong,frapresult,t_array,.03,kcompl, 1,0,300, seeds[i],nribs,x0,9, pl, 2, kon, koff,k_probe,pv,flags,N_rib)
    

  
    all_results[i,:,:] = result.T
    
    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    endcolrec = np.where(colpointsx == 0)[0][0]
    
    colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
    all_col_points.append(colpoints.T)

lowmem_t = time.time()-start
print('time for {0} trajectories {1}'.format(n_trajectories,time.time()-start))

plt.plot(all_results[0],'r-')
plt.xlabel('time')
plt.ylabel('intensity')
plt.title('recording locations')
plt.legend(['low_mem','recoding_locations'])







# bursting, leaky, stats
flags = np.array([1,0,0],dtype=np.int32)
start = time.time()
for i in range(n_trajectories):
    result = np.zeros((ncolor,len(t_array)),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    
    colpointsx = np.zeros(len(kelong)*400,dtype=np.int32)
    colpointst = np.zeros(len(kelong)*400,dtype=np.float64)

    ssa_translation_lowmem.run_SSA(result,ribtimes,coltimes,colpointsx,colpointst, kelong,frapresult,t_array,.03,kcompl, 1,0,300, seeds[i],nribs,x0,9, pl, 2, kon, koff,k_probe,pv,flags,N_rib)
    

  
    all_results[i,:,:] = result.T
    
    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    endcolrec = np.where(colpointsx == 0)[0][0]
    
    colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
    all_col_points.append(colpoints.T)

bursting_lowmem_t = time.time()-start
print('time for {0} trajectories {1}'.format(n_trajectories,time.time()-start))

plt.figure(dpi=300)
plt.plot(all_results[0])
plt.xlabel('time')
plt.ylabel('intensity')
plt.title('bursting kon/koff = 3/3,  lowmemory')
plt.legend(['c1'])




flags = np.array([0,1,0],dtype=np.int32)
start = time.time()
for i in range(n_trajectories):
    result = np.zeros((ncolor,len(t_array)),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    
    colpointsx = np.zeros(len(kelong)*400,dtype=np.int32)
    colpointst = np.zeros(len(kelong)*400,dtype=np.float64)

    ssa_translation_lowmem.run_SSA(result,ribtimes,coltimes,colpointsx,colpointst, kelong,frapresult,t_array,.03,kcompl, 1,0,300, seeds[i],nribs,x0,9, pl, 2, kon, koff,k_probe,pv,flags,N_rib)
    

  
    all_results[i,:,:] = result.T
    
    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    endcolrec = np.where(colpointsx == 0)[0][0]
    
    colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
    all_col_points.append(colpoints.T)

leaky_lowmem_t = time.time()-start
print('time for {0} trajectories {1}'.format(n_trajectories,time.time()-start))

plt.figure(dpi=300)
plt.plot(all_results[0])
plt.xlabel('time')
plt.ylabel('intensity')
plt.title('leaky,  lowmemory')
plt.legend(['c1'])








flags = np.array([1,1,0],dtype=np.int32)
start = time.time()
for i in range(n_trajectories):
    result = np.zeros((ncolor,len(t_array)),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    
    colpointsx = np.zeros(len(kelong)*400,dtype=np.int32)
    colpointst = np.zeros(len(kelong)*400,dtype=np.float64)

    ssa_translation_lowmem.run_SSA(result,ribtimes,coltimes,colpointsx,colpointst, kelong,frapresult,t_array,.03,kcompl, 1,0,300, seeds[i],nribs,x0,9, pl, 2, kon, koff,k_probe,pv,flags,N_rib)
    

  
    all_results[i,:,:] = result.T
    
    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    endcolrec = np.where(colpointsx == 0)[0][0]
    
    colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
    all_col_points.append(colpoints.T)

leaky_and_bursting_lowmem_t = time.time()-start
print('time for {0} trajectories {1}'.format(n_trajectories,time.time()-start))

plt.figure(dpi=300)
plt.plot(all_results[0])
plt.xlabel('time')
plt.ylabel('intensity')
plt.title('leaky and bursting,  lowmemory')
plt.legend(['c1'])











flags = np.array([0,0,1],dtype=np.int32)
start = time.time()
for i in range(n_trajectories):
    result = np.zeros((ncolor,len(t_array)),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    
    colpointsx = np.zeros(len(kelong)*400,dtype=np.int32)
    colpointst = np.zeros(len(kelong)*400,dtype=np.float64)

    ssa_translation_lowmem.run_SSA(result,ribtimes,coltimes,colpointsx,colpointst, kelong,frapresult,t_array,.03,kcompl, 1,0,300, seeds[i],nribs,x0,9, pl, 2, kon, koff,k_probe,pv,flags,N_rib)
    

  
    all_results[i,:,:] = result.T
    
    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    endcolrec = np.where(colpointsx == 0)[0][0]
    
    colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
    all_col_points.append(colpoints.T)

stats_lowmem = time.time()-start
print('time for {0} trajectories {1}'.format(n_trajectories,time.time()-start))

plt.figure(dpi=300)
plt.hist(ribtimes[ribtimes > 0])
plt.xlabel('time')
plt.ylabel('intensity')
plt.title('ribosometimes,  lowmemory')
plt.legend(['c1','c2'])


times = [full_rec_t,lowmem_t,bursting_lowmem_t,leaky_lowmem_t,leaky_and_bursting_lowmem_t, stats_lowmem]
plt.figure(dpi=300)
plt.bar(['full','lowmem','bursting','leaky','bursting/leaky','stats'],times)



try:
    from setuptools import find_packages
    from setuptools import setup
    from setuptools import Extension
    
except ImportError:
    from distutils.core import setup
    from distutils.extension import Extension
    
from Cython.Distutils import build_ext
from Cython.Build import cythonize

import sysconfig
import os

# Call cythonize in advance so a single module can be compiled from a single Cython
# file along with other C++ files.
import numpy as np


with open('README.md','r') as f:
    long_description = f.read()

#sources = ['ssa_translation.pyx','ssa_translation_c_w.cpp','ssa_translation_lowmem.pyx','ssa_translation_c_w_lowmem.cpp','ssa_translation_lowmem_leaky.pyx','ssa_translation_c_w_lowmem_leaky.cpp','ssa_translation_lowmem_nostats.pyx','ssa_translation_c_w_lowmem_nostats.cpp','ssa_translation_lowmem_leaky_nostats.pyx','ssa_translation_c_w_lowmem_leaky_nostats.cpp','ssa_translation_lowmem_bursting.pyx','ssa_translation_c_w_lowmem_bursting.cpp']
sources = ['ssa_translation_c_w_full.cpp','ssa_translation_lowmem.pyx','ssa_translation_c_w_lowmem.cpp',  'ssa_translation_generic_lowmem_c_w.cpp', 'ssa_translation_generic_c_w.cpp',  'ssa_trna_lowmem_c_w.cpp', 'ssa_trna_c_w.cpp']

cythonize('*.pyx', language='c++')


if not sysconfig.get_config_var('LIBS'):
    libs = sysconfig.get_config_var('LIBS')
else:
    libs = []

#try:
#
#    setup(name='SSA',ext_modules=[Extension('ssa_translation', sources, language='c++',include_dirs = ['/usr/local/include','/Library/Developer/CommandLineTools/usr/bin','/anaconda/lib/python3.6/site-packages/numpy/core/include','/anaconda/lib/python2.7/site-packages/numpy/core/include','/anaconda/lib/python3.6/site-packages/numpy/core/include',np.get_include(),'.',os.getcwd()])],cmdclass = {'build_ext': build_ext})
#except:
#    setup(name='SSA',ext_modules=cythonize([Extension('ssa_translation', sources,language='c++',extra_compile_args=[    "-stdlib=libc++"],    include_dirs = ['usr/include','/usr/local/include','/Library/Developer/CommandLineTools/usr/bin','/anaconda/lib/python3.6/site-packages/numpy/core/include',np.get_include(),'.',os.getcwd()])]),cmdclass = {'build_ext': build_ext})
#

class DependencyError(Exception):
    pass

try:
    packages=find_packages()
except:
    pass

try:
    eca = sysconfig.get_config_var('CPPFLAGS').split()
except:
    eca = []
    
# Try to find eigen directory if possible
include_list = [sysconfig.get_paths()['include'],np.get_include(),'.', os.getcwd()]

env_location = sysconfig.get_config_vars()['prefix']
print(env_location)
potential_eigens = []
for root, dirs, files in os.walk(env_location):
    for folder in dirs:
        if 'eigen3' in folder:
			
            potential_eigens.append( os.path.dirname (os.path.join(root, folder)))

print(potential_eigens)
final_dirs = []
for directory in potential_eigens:
    for root, dirs, files in os.walk(directory):
        for file in files:
            if file == 'Dense':
                final_dirs.append(directory)

print('Located eigen3/Eigen/Dense:')
print(final_dirs)    

if len(final_dirs) > 0:
    include_list.append(final_dirs[0])
else:
    raise DependencyError("Cannot find Eigen installed on enviroment, please conda install eigen or provide a path to eigen with the setup command: python setup.py build_ext --inplace -I[path to eigen, no space after I, no brackets]")
     
setup(name='rsnapsim_ssa_cpp',
      ext_modules=[Extension('ssa_translation_lowmem', 
                sources, language='c++',
                include_dirs = include_list ,
                library_dirs = libs,
                extra_compile_args= eca)]
    ,cmdclass = {'build_ext': build_ext}
    ,author='William Raymond'
	,author_email='wsraymon@rams.colostate.edu'
    ,description= 'mRNA translation Stochastic Simulation Algorithms (SSA) for the rSNAPsim module.'
    ,version = "0.0.16"
    ,long_description = long_description
    ,long_description_content_type='text/markdown'
    ,url = 'https://github.com/MunskyGroup/rSNAPsim'
    ,install_requires = ['numpy',"Cython>=0.26.0"]
    ,packages=packages
    
    )





import numpy as np
import ssa_translation
import matplotlib.pyplot as plt
import time
import scipy.special

k = np.linspace(0,200,101,)
p = (.03*1000)**k*np.exp(-.03*1000)/(scipy.special.factorial(k))


# load the elongation 
kelong = np.loadtxt('elongationrates.txt')
kbind = kelong[0]
kcompl = kelong[-1]
kelong = kelong[1:-1]

t_array = np.array([0,100,500],dtype=np.float64)
t0 = 15
t_array = np.linspace(0,1000,1000,dtype=np.float64)
N_rib = 11
result = np.zeros((len(t_array)*N_rib),dtype=np.int32  )
#kelong = np.array([3.1,3.2,3.3,3.4,3.5,3.1,3.2,3.3,3.4,3.5],dtype=np.float64)
n_trajectories = 10
start = time.time()

#preallocated arrays here
all_results = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)
lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t_array<20)[0]))

all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)

all_ribtimes = np.zeros((n_trajectories,400),dtype=np.float64)
all_coltimes = np.zeros((n_trajectories,400),dtype=np.int32)
nribs = np.array([0],dtype=np.int32)
all_ribs = np.zeros((n_trajectories,1))
seeds = np.random.randint(0,0x7FFFFFF,n_trajectories)
x0 = np.zeros((N_rib),dtype=np.int32)


print(x0.shape)
print(all_results.shape)
print(all_frapresults.shape)
print(all_ribtimes.shape)
print(all_coltimes.shape)

all_col_points = []
for i in range(n_trajectories):
    result = np.zeros((len(t_array)*N_rib),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    
    colpointsx = np.zeros(len(kelong)*400,dtype=np.int32)
    colpointst = np.zeros(len(kelong)*400,dtype=np.float64)
    
    ssa_translation.run_SSA(result,ribtimes,coltimes,colpointsx,colpointst, kelong,frapresult,t_array,.03,kcompl, 1,0,300, seeds[i],nribs,x0,9,N_rib)
    
    

    all_results[i,:] = result
    print(all_results.shape)
    print(result.shape)
    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    endcolrec = np.where(colpointsx == 0)[0][0]
    
    colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
    all_col_points.append(colpoints.T)
print('time for {0} trajectories {1}'.format(n_trajectories,time.time()-start))
#plt.hist(result[result>0])
#plt.show()
#traj = result.reshape((N_rib,len(t_array))).T
##print('The result is \n {0}'.format(result.reshape((N_rib,len(t_array))).T))
#plt.plot(traj[-1,:])
#plt.show()

# map to fluorescence.
ntimes = len(t_array)
intensity_vec = np.zeros(ntimes)
pv = np.loadtxt('probe_design.txt')
tstart = 0
I = np.zeros((n_trajectories,ntimes-tstart))
for i in range(n_trajectories):
    traj = all_results[i,:].reshape((N_rib,len(t_array))).T
    for j in range(tstart,ntimes):
        temp_output = traj[j,:]
        I[i,j] = np.sum(pv[temp_output[temp_output>0]-1])

# Plotting
all_traj = np.loadtxt('ivec_1000t')
#all_traj = ssa_obj_01.intensity_vec
f,ax = plt.subplots(2,1)
ax[0].plot(I[0:50,-500:].T)
ax[1].plot(all_traj[0:50,-500:].T)
f2,ax2 = plt.subplots(1,2)
ax2[0].hist(I[:,-500:].ravel())
ax2[1].hist(all_traj[:100,-500:].ravel())
plt.show()




# -*- coding: utf-8 -*-
"""
Created on Fri Mar 27 18:07:30 2020

@author: willi
"""

#Test file for generalized SSA

#import rSNAPsim as rss
import numpy as np
import time
import matplotlib.pyplot as plt

import ssa_translation_lowmem

def generate_additional_ks(enters,pauses,jumps,stops,L):
    
    def frame_check_1(L,arr):        
        return (L- arr[:,1]+1)*(arr[:,1]>0) + L*(arr[:,1]>1)
    
    def frame_check_3(L,arr):        
        return (L- arr[:,3]+1)*(arr[:,3]>0) + L*(arr[:,3]>1)            
                
    def gen_ks_1_loc(L,arr):
        arr[:,0] = arr[:,0]+frame_check_1(L,arr)
        arr[:,1] = arr[:,2]    
        arr = arr[:,0:2]
        max_arr = np.max( arr[:,0])     
        return arr,max_arr
    
    def gen_ks_3_loc(L,arr):
        arr[:,0] = arr[:,0]+ frame_check_1(L,arr)     
        arr[:,1] = arr[:,2]+ frame_check_3(L,arr)
        arr[:,2] = arr[:,4]
        arr = arr[:,0:3]
        max_arr = max([np.max( arr[:,0]),np.max( arr[:,1])])
        return arr,max_arr

    max_enter = 0
    max_pause = 0
    max_stop = 0
    max_jump = 0
    k_jumps = np.copy(jumps)
    k_pauses = np.copy(pauses)
    k_stops = np.copy(stops)
    k_enters = np.copy(enters)
        
    
    if len(k_enters) != 0:
        k_enters,max_enter = gen_ks_1_loc(L,k_enters)

    if len(k_pauses) != 0:
        k_pauses,max_pause = gen_ks_1_loc(L,k_pauses)

    if len(k_stops) != 0:
        k_stops,max_stop = gen_ks_1_loc(L,k_stops)
    
    if len(k_jumps) != 0:
        k_jumps,max_jump = gen_ks_3_loc(L,k_jumps)
        
    max_loc = max(max_jump,max_stop,max_pause,max_enter)
    
    if max_loc <=L: 
        frames_used = 0
    if max_loc > L:
        frames_used = 1
    if max_loc > 2*L-1 :
        frames_used = 2
    
    return k_enters, k_pauses, k_stops, k_jumps, frames_used



#rsnap = rss.rSNAPsim()
#rsnap.open_seq_file('gene_files/H2B_withTags.txt')
#rsnap.run_default()

k = np.ones((1,300)).flatten()

kelong = k[1:-1]
kelong[49] = 0
kelong[149]= 0

kelong[248] = 0

#k_fss = np.array([[200,0,200,1,.3]])
k_pause = np.array([[30,0,.005]])
k_enters = np.array([[5,0,.02],[5,2,.04]],dtype=np.float64)
k_stops = np.array([[50,0,10],[50,1,10],[50,2,10]],dtype=np.float64)
k_fss = np.array([[20,0,20,1,1]],dtype=np.float64)
#k_pause =  np.array([[30,2,100],[40,2,100]],dtype=np.float64)

k_enters,k_pauses,k_stops,k_jumps,frames_used = generate_additional_ks(k_enters,[],k_fss,k_stops,100)




t_array = np.array([0,100,500],dtype=np.float64)
t0 = 15
t_array = np.linspace(0,400,400,dtype=np.float64)
N_rib = 200
result = np.zeros((len(t_array)*N_rib),dtype=np.int32  )
#kelong = np.array([3.1,3.2,3.3,3.4,3.5,3.1,3.2,3.3,3.4,3.5],dtype=np.float64)
n_trajectories = 10
start = time.time()
all_results = np.zeros((n_trajectories,2,len(t_array)),dtype=np.int32)
lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t_array<20)[0]))

all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)




all_ribtimes = np.zeros((n_trajectories,400),dtype=np.float64)
all_coltimes = np.zeros((n_trajectories,400),dtype=np.int32)
nribs = np.array([0],dtype=np.int32)
all_ribs = np.zeros((n_trajectories,1))
seeds = np.random.randint(0,0x7FFFFFF,n_trajectories)


k_add = np.hstack((k_enters.flatten(),k_pauses.flatten(),k_stops.flatten(),k_jumps.flatten() ))
t_array_copy = np.copy(t_array)
while t_array_copy.shape[0] != 200:
    t_array_copy = np.vstack((t_array_copy,t_array))
 

probe = np.zeros((298,2)).T

probe[0,10] = 1
probe[0,20] = 1
probe[1,225] = 1
probe[1,215] = 1
#probe = np.cumsum(probe,axis=1)
probe = probe.astype(int).copy(order='C')
for i in range(n_trajectories):
    result = np.zeros((2,len(t_array)),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)



    



    ssa_translation_lowmem.run_SSA_generic_lowmem(result, kelong,frapresult,t_array, np.array([0,0,0],dtype=np.float64), seeds[i], k_add.flatten()  ,2,0,3,1, probe, 2,200 )
    all_results[i,:,:] = result
    all_frapresults[i,:] = frapresult


print(result.shape) 
print(kelong.shape),
print(frapresult.shape)

print(k_add.flatten().shape)
print(probe.T.astype(int).shape)


traj = all_results[0,:,:].reshape((2,len(t_array))).T

f,ax = plt.subplots(2,1)    

ax[0].set_ylim([0,300])
ax[0].fill_between([0,400],[100,100],color='red',alpha=.2)
ax[0].fill_between([0,400],[200,200],color='green',alpha=.2)
ax[0].fill_between([0,400],[300,300],color='blue',alpha=.2)

ax[0].plot(traj,'.')
ax[0].set_xlabel('Time')
ax[0].set_ylabel('Ribosome Location')
ax[0].set_title(' 100 codons,  enters: 0,10 and +2,10  FSS: 0,20 to +1,20 Stops: 50 0,1,2'  )

spatial_x = (traj + (traj > 100) + (traj > 199))%100

ax[1].set_ylim([0,100])


#ax[1].plot(t_array,spatial_x,'.')

ax[1].plot(t_array_copy.T[traj<=100],spatial_x[traj <= 100],'r.')
ax[1].plot(t_array_copy.T[traj>100],spatial_x[traj > 100],'g.')

ax[1].plot(t_array_copy.T[traj>199],spatial_x[traj > 199],'b.')

ax[1].set_xlabel('Time')
ax[1].set_ylabel('Ribosome Location')
ax[1].set_title(' spatial location '  )
ax[1].legend(['0','+1','+2'])






1/0





###################################################################

k = np.ones((1,300)).flatten()

kelong = k[1:-1]
kelong[49] = 3
kelong[79] = 0

k_enters = np.array([[10,0,.04]],dtype=np.float64)
k_stops = np.array([[50,0,10],[80,0,10]],dtype=np.float64)
k_fss = []
k_pause = []
#k_pause =  np.array([[30,2,100],[40,2,100]],dtype=np.float64)

k_enters,k_pauses,k_stops,k_jumps,frames_used = generate_additional_ks(k_enters,k_pause,k_fss,k_stops,100)




t_array = np.array([0,100,500],dtype=np.float64)
t0 = 15
t_array = np.linspace(0,400,400,dtype=np.float64)
N_rib = 200
result = np.zeros((len(t_array)*N_rib),dtype=np.int32  )
#kelong = np.array([3.1,3.2,3.3,3.4,3.5,3.1,3.2,3.3,3.4,3.5],dtype=np.float64)
n_trajectories = 1
start = time.time()
all_results = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)
lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t_array<20)[0]))

all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)




all_ribtimes = np.zeros((n_trajectories,400),dtype=np.float64)
all_coltimes = np.zeros((n_trajectories,400),dtype=np.int32)
nribs = np.array([0],dtype=np.int32)
all_ribs = np.zeros((n_trajectories,1))
seeds = np.random.randint(0,0x7FFFFFF,n_trajectories)


k_add = np.hstack((k_enters.flatten(),k_pauses.flatten(),k_stops.flatten(),k_jumps.flatten() ))
t_array_copy = np.copy(t_array)
while t_array_copy.shape[0] != 200:
    t_array_copy = np.vstack((t_array_copy,t_array))
 


for i in range(n_trajectories):
    result = np.zeros((len(t_array)*N_rib),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    ssa_translation_lowmem.run_SSA_generic(result,ribtimes,coltimes, kelong,frapresult,t_array, np.array([0,0,0],dtype=np.float64), seeds[i],nribs, k_add.flatten()  ,len(k_enters),len(k_pauses),len(k_stops),len(k_jumps) )
    all_results[i,:] = result
    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    

traj = all_results[0,:].reshape((N_rib,len(t_array))).T

f,ax = plt.subplots(2,1)    

ax[0].set_ylim([0,300])
ax[0].fill_between([0,400],[100,100],color='red',alpha=.2)
ax[0].fill_between([0,400],[200,200],color='green',alpha=.2)
ax[0].fill_between([0,400],[300,300],color='blue',alpha=.2)

ax[0].plot(traj,'.')
ax[0].set_xlabel('Time')
ax[0].set_ylabel('Ribosome Location')
ax[0].set_title(' 100 codons,  enters: 0,10 stops: 0,50 and 0,80'  )

spatial_x = (traj + (traj > 100) + (traj > 199))%100

ax[1].set_ylim([0,100])


#ax[1].plot(t_array,spatial_x,'.')

ax[1].plot(t_array_copy.T[traj<=100],spatial_x[traj <= 100],'r.')
ax[1].plot(t_array_copy.T[traj>100],spatial_x[traj > 100],'g.')

ax[1].plot(t_array_copy.T[traj>199],spatial_x[traj > 199],'b.')

ax[1].set_xlabel('Time')
ax[1].set_ylabel('Ribosome Location')
ax[1].set_title(' spatial location '  )
ax[1].legend(['0','+1','+2'])




#####################################################
k = np.ones((1,300)).flatten()

kelong = k[1:-1]

kelong[49] = 0
kelong[179] = 0

k_enters = np.array([[10,0,.04]],dtype=np.float64)
k_stops = np.array([[50,0,10],[80,1,10]],dtype=np.float64)
k_fss = np.array([[30,0,30,1,1]],dtype=np.float64)
k_pause = []
#k_pause =  np.array([[30,2,100],[40,2,100]],dtype=np.float64)

k_enters,k_pauses,k_stops,k_jumps,frames_used = generate_additional_ks(k_enters,k_pause,k_fss,k_stops,100)




t_array = np.array([0,100,500],dtype=np.float64)
t0 = 15
t_array = np.linspace(0,400,400,dtype=np.float64)
N_rib = 200
result = np.zeros((len(t_array)*N_rib),dtype=np.int32  )
#kelong = np.array([3.1,3.2,3.3,3.4,3.5,3.1,3.2,3.3,3.4,3.5],dtype=np.float64)
n_trajectories = 1
start = time.time()
all_results = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)
lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t_array<20)[0]))

all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)




all_ribtimes = np.zeros((n_trajectories,400),dtype=np.float64)
all_coltimes = np.zeros((n_trajectories,400),dtype=np.int32)
nribs = np.array([0],dtype=np.int32)
all_ribs = np.zeros((n_trajectories,1))
seeds = np.random.randint(0,0x7FFFFFF,n_trajectories)


k_add = np.hstack((k_enters.flatten(),k_pauses.flatten(),k_stops.flatten(),k_jumps.flatten() ))
t_array_copy = np.copy(t_array)
while t_array_copy.shape[0] != 200:
    t_array_copy = np.vstack((t_array_copy,t_array))
 


for i in range(n_trajectories):
    result = np.zeros((len(t_array)*N_rib),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    ssa_translation_lowmem.run_SSA_generic(result,ribtimes,coltimes, kelong,frapresult,t_array, np.array([0,0,0],dtype=np.float64), seeds[i],nribs, k_add.flatten()  ,1,0,2,1 )
    all_results[i,:] = result
    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    

traj = all_results[0,:].reshape((N_rib,len(t_array))).T

f,ax = plt.subplots(2,1)    

ax[0].set_ylim([0,300])
ax[0].fill_between([0,400],[100,100],color='red',alpha=.2)
ax[0].fill_between([0,400],[200,200],color='green',alpha=.2)
ax[0].fill_between([0,400],[300,300],color='blue',alpha=.2)

ax[0].plot(traj,'.')
ax[0].set_xlabel('Time')
ax[0].set_ylabel('Ribosome Location')
ax[0].set_title(' 100 codons,  enters: 0,10 stops: 0,50 and 1,80  FSS: 0,30 to 1,30'  )

spatial_x = (traj + (traj > 100) + (traj > 199))%100

ax[1].set_ylim([0,100])


#ax[1].plot(t_array,spatial_x,'.')

ax[1].plot(t_array_copy.T[traj<=100],spatial_x[traj <= 100],'r.')
ax[1].plot(t_array_copy.T[traj>100],spatial_x[traj > 100],'g.')

ax[1].plot(t_array_copy.T[traj>199],spatial_x[traj > 199],'b.')

ax[1].set_xlabel('Time')
ax[1].set_ylabel('Ribosome Location')
ax[1].set_title(' spatial location '  )
ax[1].legend(['0','+1','+2'])




######################

k = np.ones((1,300)).flatten()

kelong = k[1:-1]

kelong[49] = 0
kelong[278] = 0

k_enters = np.array([[10,0,.04],[10,2,.02]],dtype=np.float64)
k_stops = np.array([[50,0,10],[80,2,10]],dtype=np.float64)
k_fss = []
k_pause = []
#k_pause =  np.array([[30,2,100],[40,2,100]],dtype=np.float64)

k_enters,k_pauses,k_stops,k_jumps,frames_used = generate_additional_ks(k_enters,k_pause,k_fss,k_stops,100)




t_array = np.array([0,100,500],dtype=np.float64)
t0 = 15
t_array = np.linspace(0,400,400,dtype=np.float64)
N_rib = 200
result = np.zeros((len(t_array)*N_rib),dtype=np.int32  )
#kelong = np.array([3.1,3.2,3.3,3.4,3.5,3.1,3.2,3.3,3.4,3.5],dtype=np.float64)
n_trajectories = 1
start = time.time()
all_results = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)
lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t_array<20)[0]))

all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)




all_ribtimes = np.zeros((n_trajectories,400),dtype=np.float64)
all_coltimes = np.zeros((n_trajectories,400),dtype=np.int32)
nribs = np.array([0],dtype=np.int32)
all_ribs = np.zeros((n_trajectories,1))
seeds = np.random.randint(0,0x7FFFFFF,n_trajectories)


k_add = np.hstack((k_enters.flatten(),k_pauses.flatten(),k_stops.flatten(),k_jumps.flatten() ))
t_array_copy = np.copy(t_array)
while t_array_copy.shape[0] != 200:
    t_array_copy = np.vstack((t_array_copy,t_array))
 


for i in range(n_trajectories):
    result = np.zeros((len(t_array)*N_rib),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    ssa_translation_lowmem.run_SSA_generic(result,ribtimes,coltimes, kelong,frapresult,t_array, np.array([0,0,0],dtype=np.float64), seeds[i],nribs, k_add.flatten()  ,2,0,2,0 )
    all_results[i,:] = result
    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    

traj = all_results[0,:].reshape((N_rib,len(t_array))).T

f,ax = plt.subplots(2,1)    

ax[0].set_ylim([0,300])
ax[0].fill_between([0,400],[100,100],color='red',alpha=.2)
ax[0].fill_between([0,400],[200,200],color='green',alpha=.2)
ax[0].fill_between([0,400],[300,300],color='blue',alpha=.2)

ax[0].plot(traj,'.')
ax[0].set_xlabel('Time')
ax[0].set_ylabel('Ribosome Location')
ax[0].set_title(' 100 codons,  enters: 0,10 2,20 stops: 0,50 and 2,80'  )

spatial_x = (traj + (traj > 100) + (traj > 199))%100

ax[1].set_ylim([0,100])


#ax[1].plot(t_array,spatial_x,'.')

ax[1].plot(t_array_copy.T[traj<=100],spatial_x[traj <= 100],'r.')
ax[1].plot(t_array_copy.T[traj>100],spatial_x[traj > 100],'g.')

ax[1].plot(t_array_copy.T[traj>199],spatial_x[traj > 199],'b.')

ax[1].set_xlabel('Time')
ax[1].set_ylabel('Ribosome Location')
ax[1].set_title(' spatial location '  )
ax[1].legend(['0','+1','+2'])








###########

k = np.ones((1,300)).flatten()

kelong = k[1:-1]

kelong[49] = 0

kelong[39] = 0.1
kelong[278] = 0

k_enters = np.array([[10,0,.04],[10,2,.02]],dtype=np.float64)
k_stops = np.array([[50,0,10],[80,2,10]],dtype=np.float64)
k_fss = []
k_pause =  np.array([[40,0,100]],dtype=np.float64)
#k_pause =  np.array([[30,2,100],[40,2,100]],dtype=np.float64)

k_enters,k_pauses,k_stops,k_jumps,frames_used = generate_additional_ks(k_enters,k_pause,k_fss,k_stops,100)




t_array = np.array([0,100,500],dtype=np.float64)
t0 = 15
t_array = np.linspace(0,400,400,dtype=np.float64)
N_rib = 200
result = np.zeros((len(t_array)*N_rib),dtype=np.int32  )
#kelong = np.array([3.1,3.2,3.3,3.4,3.5,3.1,3.2,3.3,3.4,3.5],dtype=np.float64)
n_trajectories = 1
start = time.time()
all_results = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)
lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t_array<20)[0]))

all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)




all_ribtimes = np.zeros((n_trajectories,400),dtype=np.float64)
all_coltimes = np.zeros((n_trajectories,400),dtype=np.int32)
nribs = np.array([0],dtype=np.int32)
all_ribs = np.zeros((n_trajectories,1))
seeds = np.random.randint(0,0x7FFFFFF,n_trajectories)


k_add = np.hstack((k_enters.flatten(),k_pauses.flatten(),k_stops.flatten(),k_jumps.flatten() ))
t_array_copy = np.copy(t_array)
while t_array_copy.shape[0] != 200:
    t_array_copy = np.vstack((t_array_copy,t_array))
 


for i in range(n_trajectories):
    result = np.zeros((len(t_array)*N_rib),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    ssa_translation_lowmem.run_SSA_generic(result,ribtimes,coltimes, kelong,frapresult,t_array, np.array([0,0,0],dtype=np.float64), seeds[i],nribs, k_add.flatten()  ,len(k_enters),len(k_pauses),len(k_stops),len(k_jumps))
    all_results[i,:] = result
    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    

traj = all_results[0,:].reshape((N_rib,len(t_array))).T

f,ax = plt.subplots(2,1)    

ax[0].set_ylim([0,300])
ax[0].fill_between([0,400],[100,100],color='red',alpha=.2)
ax[0].fill_between([0,400],[200,200],color='green',alpha=.2)
ax[0].fill_between([0,400],[300,300],color='blue',alpha=.2)

ax[0].plot(traj,'.')
ax[0].set_xlabel('Time')
ax[0].set_ylabel('Ribosome Location')
ax[0].set_title(' 100 codons,  enters: 0,10 2,20 stops: 0,50 and 2,80'  )

spatial_x = (traj + (traj > 100) + (traj > 199))%100

ax[1].set_ylim([0,100])


#ax[1].plot(t_array,spatial_x,'.')

ax[1].plot(t_array_copy.T[traj<=100],spatial_x[traj <= 100],'r.')
ax[1].plot(t_array_copy.T[traj>100],spatial_x[traj > 100],'g.')

ax[1].plot(t_array_copy.T[traj>199],spatial_x[traj > 199],'b.')

ax[1].set_xlabel('Time')
ax[1].set_ylabel('Ribosome Location')
ax[1].set_title(' spatial location '  )
ax[1].legend(['0','+1','+2'])






# -*- coding: utf-8 -*-
"""
Created on Tue Jun 22 15:43:07 2021

@author: willi
"""
import numpy as np
import rsnapsim as rsim
h2b_gene_file = '../gene_files/.txt'

poi_seqs, pois, tagged_pois, seq = rsim.seqmanip.open_seq_file(h2b_gene_file)
h2b_construct = pois['1'][0]
h2b_construct.generate_3frame_tags()



model = rsim.model_builder
model.add_k(h2b_construct)

model._poi.multiframe_epitopes[0]['T_Flag'] = [310, 320, 330,340,350]
model._poi.multiframe_epitopes[1]['T_New'] = [ 240,250,230,260,270]
model._poi.multiframe_epitopes[0]['T_Hemagglutinin']= [100,110,120,130,140, ]
model._poi.multiframe_epitopes[2]['T_New2']=  [310, 320, 330,340,350]


model.add_enters([5,50], [0,1], [.08,.03])


model.add_jumps([299], [1], [300], [2], [2])
model.visualize_transcript()

t_array = np.linspace(0,1000,1001,dtype=np.float64)
len(h2b_construct.kelong)
#r = model.run_ssa(t_array,n_traj=10,low_mem=False)


# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 17:57:33 2020

@author: willi
"""


import json, codecs
from json import encoder
import numpy as np
from . import GenericMetaData
GenericMetaData = GenericMetaData.GenericMetaData


class SSA_Soln():
    '''
    SSA container class

    holds intensity / ribosome data as well as the propensities used

    __.n_traj = number of trajectories
    __.k = propensities used for the simulation
    __.rib_density = ribosome density per mRNA strand
    __.ribosome_means


    '''
    def __init__(self):
        self.n_traj = 0   #number trajectories
        self.k = []       #propensities
        self.no_rib_per_mrna = 0    #number of ribosomes per mrna strand
        self.rib_density = 0      #ribosome density
        self.ribosome_means = 0  #mean ribosomes
        self.rib_vec = 0          #actual vector of ribosome locations
        self.intensity_vec = []   #intensity vectors per SSA trajectory
        self.time_vec_fixed = []   #time array
        self.__meta = GenericMetaData().get()



    def save(self, filename, precision='.4f'):
        ext = filename.split('.')[-1]

        if ext == 'txt':
            self.__save_txt(filename)
        if ext == 'json':
            self.__save_json(filename, precision=precision)


    def load(self, filename):
        ext = filename.split('.')[-1]

        if ext == 'txt':
            self.__load_from_txt(filename)
        if ext == 'json':
            self.__load_from_json(filename)

    def __save_txt(self, filename):

        if '.txt' in filename:
            filehandle = open(filename, 'a')
            for key in self.__dict__.keys():

                if key != 'rib_vec' and key != 'ribosome_means':
                    filehandle.write((key + '\r\n'))
                    np.savetxt(filehandle, np.atleast_2d(self.__dict__[key]),
                               delimiter=',', fmt='%s')
                    filehandle.write(('\r\n'))

        else:
            filename = filename + '.txt'
            filehandle = open(filename, 'a')
            for key in self.__dict__.keys():

                if key != 'rib_vec' and key != 'ribosome_means':
                    filehandle.write((key + '\r\n'))
                    np.savetxt(filehandle, np.atleast_2d(self.__dict__[key]),
                               delimiter=',', fmt='%s')
                    filehandle.write(('\r\n'))
        filehandle.close()


    def __load_from_txt(self, filename):
        if '.txt' in filename:
            ssa_obj = np.loadtxt(filename, dtype=str, delimiter='\n')
            solutions = []
            for i in range(0, len(ssa_obj)-1):
                label = ssa_obj[i]


                if label in ['rib_means',
                             'rib_vec',
                             'n_traj',
                             'start_time',
                             'k',
                             'time_vec_fixed',
                             'dwelltime',
                             'mean_autocorr',
                             'no_rib_per_mrna',
                             'ke_sim',
                             'autocorr_vec',
                             'ribosome_means',
                             'error_autocorr',
                             'rib_density',
                             'time',
                             'ke_true',
                             'evaluating_inhibitor',
                             'time_inhibit',
                             'evaluating_frap']:

                    if label in ['start_time', 'no_rib_per_mrna',
                                 'ke_sim', 'dwelltime', 'ke_true',
                                 'time_inhibit']:

                        array = np.fromstring(ssa_obj[i+1], dtype=float, sep=',')[0]
                        exec(('self.'+label+ '=array'))
                    elif label in ['n_traj']:
                        array = int(np.fromstring(ssa_obj[i+1], dtype=float, sep=',')[0])
                        exec(('self.'+label+ '=array'))
                    else:
                        array = np.fromstring(ssa_obj[i+1], dtype=float, sep=',')
                        exec(('self.'+label+ '=array'))

                if label in ['evaluating_inhibitor', 'evaluating_frap']:
                    if 'False' in ssa_obj[i+1]:
                        exec(('self.'+label+ '=False'))
                    if 'True' in ssa_obj[i+1]:
                        exec(('self.'+label+ '=True'))


            for i in range(0, len(ssa_obj)-1):
                label = ssa_obj[i]

                if label == 'intensity_vec':

                    #tvec = self.time_vec_fixed[np.where(self.time_vec_fixed >= self.start_time)]
                    i_vec = np.zeros((self.n_traj, len(self.time)))

                    for j in range(self.n_traj):
                        array = np.fromstring(ssa_obj[i+j+1],
                                              dtype=float, sep=',')
                        i_vec[j] = array

                    exec(('self.'+label+ '=i_vec'))

                if label == 'solutions':
                    for j in range(self.n_traj):
                        array = np.fromstring(ssa_obj[i+j+1],
                                              dtype=float, sep=',')
                        solutions.append(array)

                    exec(('self.'+label+ '=solutions'))




    def make_dict(self):
        ssadict = {}
        for key in self.__dict__.keys():
            print(key)
            if key != 'rib_vec' and key != 'ribosome_means':
                try:
                    ssadict[key] = self.__dict__[key].tolist()
                except:
                    ssadict[key] = self.__dict__[key]

            if key == 'col_points':
                col_pt = [x.tolist() for x in self.__dict__[key]]
                ssadict[key] = col_pt


        return ssadict


    def __set_float_pres(self, precision='.4f'):
        print(precision)
        encoder.FLOAT_REPR = lambda o: format(o, precision)


    def __make_float(self, float_like, precision):
        return float(('%' + precision) % float_like)

    def __format_floats(self, arraylike, precision='.4f'):

        n_decimals = int(precision.split('.')[1][:-1])
        if isinstance(arraylike, np.ndarray):
            tmp_arr = np.around(arraylike, decimals=n_decimals) .tolist()
            return tmp_arr
        else:
            tmp_arr = arraylike
            if isinstance(tmp_arr, float):
                return self.__make_float(tmp_arr, precision)
            if isinstance(tmp_arr, list):
                return [self.__make_float(x, precision) for x in tmp_arr]


    def __save_json(self, filename, precision='.4f'):

        if '.json' in filename:

            ssadict = {}
            for key in self.__dict__.keys():

                if key != 'col_points':

                    if type(self.__dict__[key]) in [float, list, np.ndarray]:

                        ssadict[key] = self.__format_floats(self.__dict__[key],
                                                            precision=precision)
                    else:
                        ssadict[key] = self.__dict__[key]

                if key == 'col_points':
                    col_pt = [x.tolist() for x in self.__dict__[key]]
                    ssadict[key] = col_pt

            json.dump(
                ssadict, codecs.open(
                    filename, 'w', encoding='utf-8'),
                separators=(',', ':'), sort_keys=True, indent=4)

        else:
            filename = filename + '.json'

            ssadict = {}
            for key in self.__dict__.keys():
                if key != 'rib_vec' and key != 'ribosome_means':
                    try:
                        ssadict[key] = self.ssa_harr.__dict__[key].tolist()
                    except:
                        ssadict[key] = self.ssa_harr.__dict__[key]

            json.dump(
                ssadict, codecs.open(
                    filename, 'w', encoding='utf-8'),
                separators=(',', ':'), sort_keys=True, indent=4)


    def __load_from_json(self, filename):
        if '.json' in filename:

            obj_text = codecs.open(filename, 'r', encoding='utf-8').read()
            ssadict = json.loads(obj_text)

            for key in ssadict.keys():
                if key in ['solutions', 'all_trna_results',
                           'rib_means', 'time_vec_fixed',
                           'mean_autocorr', 'autocorr_vec',
                           'error_autocorr', 'rib_density',
                           'intensity_vec', 'I']:

                    self.__dict__[key] = np.array(ssadict[key])

                elif key in ['colpoints']:

                    cpts = [np.array(x) for x in ssadict[key]]
                    self.__dict__[key] = cpts
                else:
                    self.__dict__[key] = ssadict[key]



# -*- coding: utf-8 -*-
"""
Created on Mon May  3 12:38:15 2021

@author: willi
"""

import numpy as np
import itertools

class tRNAmap():
    def __init__(self):
        self.trna_map = None
        
    
    
    
        # https://www.nature.com/articles/nsmb866
        #5' anticodon base to 3' codon base
        self.bp_rules = {'A':'U', 'G':'C','C':'G','U':'A'}
        
        self.bp_wobble_rules = {'A':['U','C','G','A'],
                                'C':['G'],
                                'G':['C','U'],
                                'U':['A','G','C','U'],
                                'I':['A','C','U']}
        
        self.hg19_trna_gene_number_anticodon = {'AGC': 33, 'CGC':5,
                                             'UGC':11, 'ACG': 8, 'GGC':2,
                                             'CCG':4, 'CCU': 8,
                                             'UCG':6, 'UCU': 6,
                                             'AUU':2, 'GUU':36,
                                             'GUC':19, 'GCA':36,
                                             'CUG':21,'UUG':10,
                                             'CUC':13, 'UUC':14,
                                             'CCC':13, 'GCC':15,
                                             'UCC': 11, 'GUG':10,
                                             'AAU':18, 'GAU':6, 
                                             'UAU':5, 'AAG':13,
                                             'CAA':8, 'CAG':10,
                                             'UAA':6,'UAG':4,
                                             'CUU':26,'UUU':20,
                                             'CAU':11,'GAA':15,
                                             'AGG':11, 'CGG':4,
                                             'UGG':9,'UCA':2,
                                             'AGA':11,'CGA':4,
                                             'GCU':9,'UGA':5,
                                             'CUA':0,'UUA':2,
                                             'AGU':10,'CGU':6,
                                             'UGU':6,'CCA':8,
                                             'AUA':1,'GUA':16,
                                             'AAC':12,'CAC':18, #duplicate for doubled up codons
                                             'UAC':7, }
               
    
    def get_codon_species_dict(self,trna_gene_dict, alt_stopcodons = None):
        bases = ['A','G','C','U']
        all_3mers = [''.join(i) for i in itertools.product(bases, repeat = 3)]
        if alt_stopcodons == None:
            alt_stopcodons = ['UAG','UGA','UAA']
        
        for stop in alt_stopcodons:
            all_3mers.remove(stop)
            
    
        available_genes = sorted(list(trna_gene_dict.keys()))
        
        #get the available tRNA
        #available_genes = available_genes + [ 'IAC', 'IGU', 'ICG', 'IAG','IAU','IGC' ]
        ## check the anticodon orientation                                              
        try: 
            tmp_keys = list(trna_gene_dict.keys())  #get the trna dictionary keys
            tmp_keys_lower = [x.lower() for x in tmp_keys] #get the lower strings of them
            ala_key = tmp_keys[tmp_keys_lower.index('ala')] #get the ALA key
            alanines = trna_gene_dict[ala_key]
            if len(set([x[0] for x in alanines])) == 1: # if they are oriented 3' to 5' ie all starting with C's 
                flipped = False
                anti_codon_3rd_nt = 2
            elif len(set([x[0] for x in alanines])) == 3:
                flipped = True
                anti_codon_3rd_nt = 0
            else:
                ValueError('could not recognize what this dictionary is')
        except:
            flipped= True
        
        codon_to_species = {}
        for i in range(len(all_3mers)):
            codon =  all_3mers[i]
            match_ids = []
            for anticodon in available_genes:
                if flipped:
                    matches = self.generate_matches(anticodon[::-1])
                    if codon in matches:
                        match_ids =match_ids + [available_genes.index(anticodon)]
                else:
                    matches = self.generate_matches(anticodon)
                    if codon in matches:
                        match_ids =match_ids + [available_genes.index(anticodon)]
                codon_to_species[codon] = match_ids
                
            codon_to_species[codon] = match_ids
        return codon_to_species,available_genes
    
    
    def generate_map_matrix(self,trna_gene_dict, alt_stopcodons = None):
        codon_to_species, available_anticodons  =  self.get_codon_species_dict(trna_gene_dict, alt_stopcodons=alt_stopcodons )
        n_codons = len(codon_to_species)
        n_species = len(available_anticodons)
        codon_to_trna_species_map = np.zeros([n_codons,n_species])
        
        bases = ['A','G','C','U']
        all_3mers = [''.join(i) for i in itertools.product(bases, repeat = 3)]
        
        if alt_stopcodons == None:
            alt_stopcodons = ['UAG','UGA','UAA']       
        for stop in alt_stopcodons:
            all_3mers.remove(stop)
            
        for i,codon in enumerate(all_3mers):
            species = sorted(codon_to_species[codon])
            for idx in species:
                codon_to_trna_species_map[i,idx] = 1
        return codon_to_trna_species_map, all_3mers, available_anticodons
            
        
        
        
        
    def generate_matches(self,codon,rules=None):
        if rules == None:
            rules = self.bp_wobble_rules
        frnt = self.bp_rules[codon.upper()[0]] + self.bp_rules[codon.upper()[1]] 
        return [frnt + x for x in self.bp_wobble_rules[codon.upper()[2]]] 
        


# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 17:48:33 2020

@author: William Raymond
"""

import time
import sys
import platform
import rsnapsim as rSNAPsim

class GenericMetaData():
    '''
    Provides a metatdata dictionary to the user about rsnapsim files or
    models being generated.

    Attributes
    ----------

    id
    created_at
    rss_version
    platform
    python_version

    '''
    def __init__(self):
        self.id = ''
        self.created_at = time.strftime('%Y-%m-%d %H:%M:%S',
                                        time.localtime(time.time()))
        #self.user = os.path.expanduser("~")
        self.rss_version = rSNAPsim.__version__
        self.platform = platform.platform()
        self.python_version = sys.version

    def get(self):
        '''
        generate and return a metadata dictionary for a solver object

        Returns
        -------
        dict
            a dictionary of metadata such as solution id, time ran, user,
            platform and rss version.

        '''
        return self.__dict__

    def get_as_str(self):
        metadata = self.get()
        # convert to a metadata string
        meta_str = '######################################################'
        meta_str += 'model ID : ' + metadata['id'] + '\n'
        meta_str += 'files created at : ' + metadata['created_at']+ '\n'
        meta_str += 'platform : ' + metadata['platform']+ '\n'
        meta_str += 'python version : ' + metadata['python_version']+ '\n'
        meta_str += 'rsnapsim version : ' + metadata['rss_version']+ '\n'
        meta_str += '#####################################################'
        return meta_str



# -*- coding: utf-8 -*-
"""
Created on Fri Mar 27 15:15:25 2020

@author: willi
"""

import matplotlib.pyplot as plt
import time
import numpy as np
kelong = np.loadtxt('elongationrates.txt')
kbind = kelong[0]
kcompl = kelong[-1]
kelong = (np.ones((464*2,1))*3.9).flatten()

t_array = np.array([0,100,500],dtype=np.float64)
t0 = 15
t_array = np.linspace(0,1000,1000,dtype=np.float64)
N_rib = 200
result = np.zeros((len(t_array)*N_rib),dtype=np.int32  )
#kelong = np.array([3.1,3.2,3.3,3.4,3.5,3.1,3.2,3.3,3.4,3.5],dtype=np.float64)
n_trajectories = 100
start = time.time()
all_results = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)
lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t_array<20)[0]))

all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)




kenters = np.array([[0,0,kbind],[600,1,kbind]])
kjumps = np.array([[100,0, 200,1, 10]])



def generate_additional_ks(k_enters,k_pauses,k_jumps,k_stops,L):

    max_enter = 0
    max_pause = 0
    max_stop = 0
    max_jump = 0
    
    if k_enters != []:
        k_enters[:,0] = k_enters[:,0]+L*k_enters[:,1]
        k_enters[:,1] = k_enters[:,2]    
        k_enters = k_enters[:,0:2]
        max_enter = np.max( k_enters[:,0])

    if k_pauses != []:
        k_pauses[:,0] = k_pauses[:,0]+ L*k_pauses[:,1]
        k_pauses[:,1] = k_pauses[:,2]
        k_pauses = k_pauses[:,0:2]
        max_pause = np.max( k_pauses[:,0])

    if k_stops != []:
        k_stops[:,0] = k_stops[:,0]+L*k_stops[:,1]
        k_stops[:,1] = k_stops[:,2]    
        k_stops = k_stops[:,0:2]
        max_stop = np.max( k_stops[:,0])
    
    if k_jumps != []:
        k_jumps[:,0] = k_jumps[:,0]+ L*k_jumps[:,1]
        
        k_jumps[:,1] = k_jumps[:,2]+ L*k_jumps[:,3]
        k_jumps[:,2] = k_jumps[:,4]
        k_jumps = k_jumps[:,0:3]
        
        max_jump = max([np.max( k_jumps[:,0]),np.max( k_jumps[:,1])])
        
    max_loc = max(max_jump,max_stop,max_pause,max_enter)
    
    if max_loc <=L: 
        frames_used = 0
    if max_loc > L:
        frames_used = 1
    if max_loc > 2*L :
        frames_used = 1
    
    return k_enters, k_pauses, k_stops, k_jumps, frames_used
    




# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 11:49:42 2020

@author: William Raymond
"""

import numpy as np
class IntensityAnalysesRagged():
    def __init__(self):
        pass
    
    @staticmethod
    def get_acc2(data, trunc=False):
        '''
        Get autocorrelation function

        *NOT* multi-tau
        '''
        N = len(data)
        fvi = np.fft.fft(data, n=2*N)
        acf = fvi*np.conjugate(fvi)
        acf = np.fft.ifft(acf)
        acf = np.real(acf[:N])/float(N)
        if trunc:
            acf[acf < 0]=0
            for i in range(1, len(acf)):
                if acf[i] > acf[i-1]:
                    acf[i] = acf[i-1]
        return acf       
    
    def get_g0(self,covariance, mode = 'interp'):
        '''
        
        '''
        if mode.lower() in ['interp','inter','extrapolate','interpolate']:
            X = [1,2,3,4]
            G0 = np.zeros((covariance.shape[0], covariance.shape[2]))
            for i in range(covariance.shape[0]):
                for n in range(covariance.shape[0]):
                    V = covariance[i,X,n]
                    print(V.shape)
                    G0[i,n] = np.interp(0,X,V)
           
            
        if mode.lower() in ['g1','1']:
            G0 = covariance[:,1,:]
            
        if mode.lower() in ['g0','0']:
            G0 = covariance[:,0,:]

        if mode.lower() in ['max','maximum']:
            G0 = np.max(covariance,axis=1)
            
        return G0
    
    def get_autocorr(self,autocov,g0='G0'):
        '''
        normalize the autocovariance over g0
        '''
        autocorr = np.copy(autocov)
        n_traj = autocorr.shape[-1]
        autocov_error =  1.0/np.sqrt(n_traj)*np.std(autocov,ddof=1,axis=2)
        
        g0 = self.get_g0(autocov,g0)
        for n in range(autocov.shape[0]):
            autocorr[n] = autocorr[n]/g0[n]
        
        err_autocorr =  1.0/np.sqrt(n_traj)*np.std(autocorr,ddof=1,axis=2)
        return autocorr,err_autocorr
                

    def get_autocov(self,intensity_vec,max_lag,norm='raw'):
        
        colors, n_traj = self.__get_ivec_dims(intensity_vec)
        
        autocorr_vec = np.zeros((colors, max_lag, n_traj    ) )
        autocorr_err = np.zeros((colors, max_lag, n_traj  ) )
        

        
        
        for n in range(colors):
            if norm in [ 'Individual','I','individual','ind']:
                for i in range(n_traj):
                    ivec = intensity_vec[i][n]
                    print(ivec.shape)
                    trace_len = min(max_lag, len(ivec) )
                    autocorr_vec[n,:trace_len,i] = self.get_acc2( (ivec-np.mean(ivec))/np.var(ivec)  )[:trace_len]
                    
            elif norm in ['global','Global','g','G']:
                means = []
                var = []
                for i in range(n_traj):
                    ivec = intensity_vec[i][n]
                    means.append( np.mean(ivec) )
                    var.append(np.var(ivec))         
                    
                global_mean = np.mean(means)
                global_var = np.var(var)
                
                for i in range(n_traj):
                    ivec = intensity_vec[i][n]
                    trace_len = min(max_lag, len(ivec) )
                    autocorr_vec[n,:trace_len,i] = self.get_acc2((ivec-global_mean)/global_var )[:trace_len]
            elif norm in ['raw','Raw']:
                for i in range(n_traj):
                    ivec = intensity_vec[i][n]
                    trace_len = min(max_lag, len(ivec) )
                    autocorr_vec[n,:trace_len,i] = self.get_acc2(ivec)[:trace_len]  
                    
            # elif norm in ['max_ind','Max','maximum','Maximum','max']:
            #     for i in range(n_traj):
            #         ivec = intensity_vec[i][n]
            #         trace_len = min(max_lag, len(ivec) )
            #         autocorr_vec[n,:trace_len,i] = self.get_acc2(ivec)[:trace_len]
            #         autocorr_vec[n,:trace_len,i] = autocorr_vec[n,:trace_len,i]/np.max(autocorr_vec[n,:trace_len,i])
                    
            else:
                print('unrecognized normalization, please use individual, global, or none')
                return

        autocorr_err =  1.0/np.sqrt(n_traj)*np.std(autocorr_vec,ddof=1,axis=2)
                
        return autocorr_vec, autocorr_err
    
    
    def get_intensity_histogram(self,intensity_vec,n_bins,scale_factor,time_slice=1):
        colors, n_traj = self.__get_ivec_dims(intensity_vec)
        hist_list = []
        hist_edges_list = []
        for n in range(colors):
            for i in range(n_traj):
                if i == 0:
                    ivec = intensity_vec[i][n][::time_slice]
                    
                else:
                    tmp_vec = intensity_vec[i][n][::time_slice]
                    
                    ivec = np.hstack((ivec, tmp_vec))
                    
                    
            exp_int_dist_ump = np.divide(ivec,scale_factor)
            exp_int_dist_ump[exp_int_dist_ump<0] = 0
            exp_int_dist_ump[np.isnan(exp_int_dist_ump)] = 0
            
            temp_hist = np.histogram(exp_int_dist_ump, bins=n_bins)
            hist_list.append(temp_hist[1])
            hist_edges_list.append(temp_hist[0])
            
        hist_data = np.array(hist_list)
        hist_bins = np.array(hist_edges_list)   
        
        return hist_bins, hist_data
            
    
    def __get_ivec_dims(self,ivec):
        colors = ivec[0].shape[0]
        n_traj = len(ivec)           
        return colors,n_traj
     


import numpy as np
#import stochasticSimulationAlgorithm as ssa
import time
from scipy.sparse import spdiags
from scipy.linalg import solve_lyapunov
#import generic_solvers
#reload(generic_solvers)
from .generic_solvers import *


import scipy as sp
from . import expv, expvt

from scipy.sparse.linalg import expm

class TranslateModels:
    def getW1(self):
        '''
        build a propensity vector that works with my SSA. 
        '''
        W1 = np.diag(self.ke)
        f_rxn = np.zeros(self.N)
        return np.vstack((f_rxn,W1))

    def getW0(self):
        W0=np.atleast_2d(np.zeros(self.N+1)).T
        W0[0,0]=self.kb
        return W0

    def getselongation(self):
        __stoich_matrix_elongation = np.diag(np.ones(self.N),k=0)+np.diag(-1*np.ones(self.N-1),k=1)
        __f_rxn = np.atleast_2d(np.zeros(self.N))
        __f_rxn[-1,-1] = -1
        return np.hstack((__stoich_matrix_elongation,__f_rxn.T))

    def map_to_fluorescence(self,soln=None):
        self.probe_design_vector = self.get_fluor_vec()
     
        if soln is None:
            self.fluorescence = np.dot(self.probe_design_vector,self.soln)
        else:
            

            self.fluorescence = np.dot(np.dot(self.probe_design_vector,soln),self.probe_design_vector.T)
        return self.fluorescence 

    def get_fluor_vec(self,binary=None):
        if binary is not None:
            self.binary = binary
        if self.binary is None:
            n=len(self.getxi())
            return np.array(range(n))/float(n)
        else:
            return np.array([np.cumsum(self.binary,axis=1).dot(self.fi)])[0]
    
    def map_to_fluorescence2(self,color_1_bin,color_2_bin,soln):
        self.probe_design_vector_1 = self.get_fluor_vec(color_1_bin)
        self.probe_design_vector_2 = self.get_fluor_vec(color_2_bin)
        
        self.probe_design_vector = np.vstack((self.probe_design_vector_1,self.probe_design_vector_2))
        self.fluorescence = np.dot(np.dot(self.probe_design_vector,soln),self.probe_design_vector.T)
        
        return self.fluorescence   

    def map_to_fluorescence3(self,soln=None):
        '''
        Get the fluorescence at steady state
        '''
        self.probe_design_vector =  self.get_fluor_vec()
       
        #if self.soln is None:
           # print("no solution saved to memory, run self.solve() wtf")
        if soln is None:
            self.fluorescence = np.dot(self.probe_design_vector,self.soln)
        else:
            self.fluorescence = np.dot(self.probe_design_vector,soln)
        return self.fluorescence 

class TranslateODE(TranslateModels,GenericODE):
    def __init__(self,ptimes=None,tf=None,ke=None,kb=None,fi=None,N=None):
        GenericODE.__init__(self)
        if N: 
            self.ptimes=ptimes
            self.ti = 0 
            self.tf = tf
            self.params={'ke':ke,\
                         'kb':kb,\
                         'fi':fi}
            self.N=N
        else:
            self.ptimes=100
            self.ti = 0; self.tf =10.0;
            self.params={'ke':5.0,\
                         'kb':5.0,\
                         'fi':0.1}
            self.N=100
        #self.W1 = self.getW1()
        #self.W0 = self.getW0()   
        #self.S=self.getselongation()
        #self.ODE=lambda x,t:np.ravel(np.dot(self.S,(np.dot(self.W1,np.atleast_2d(x).T)))+np.dot(self.S,self.W0))
        #self.build_probe()
        #self.summary = {}
        #self.soln=None
        #self.xi=self.getxi()
    
    def solve(self):
        '''
        wrap generic run to reinitialize variables
        '''
        self.xi=np.zeros(self.N)
        self.W1=self.getW1()
        self.W0=self.getW0()
        self.S=self.getselongation()
        
        if self.N > 100:
            self.soln = self.expv_solve_means()
        
        #self.build_probe()
        else:
            self.ODE=lambda x,t:np.ravel(np.dot(np.dot(self.S,self.W1),np.atleast_2d(x).T)+np.dot(self.S,self.W0))
            solution=self._solve()
            self.soln=solution
            
        return self.map_to_fluorescence()
    
    
    def expv_solve_means(self):
        '''
        expv solver for systems that are greater than 100 states 
        
        This is specific for the Translational models as it can be written as A*X
        '''
        
        # reshape and add Kbind to the propensity and stoich, for x' = A*x
        
        S_1 = np.zeros((self.S.shape[0]+1,self.S.shape[1]))
        S_1[1:,:] = self.S
        W1_1 = np.zeros((self.W1.shape[0],self.W1.shape[1] + 1))
        
        W1_1[:,1:] = self.W1
        W1_1[0,0] = self.kb
        W1_1[0,1] = 0

        phi1 = sp.sparse.csc_matrix(np.dot(S_1,W1_1)) 
                
        x0 = np.zeros((self.N + 1,1))
        x0[0] = 1
        t = self.gettvec()
        xv =np.copy(x0) 
        #intensity = np.zeros(len(t))
        soln = np.zeros((self.N+1,len(t) ))
        
    
        for i in range(1,len(t)):
    
            xv,m,v = expv(t[i]-t[i-1],phi1,xv,tol = 1e-6,m=30) 

            soln[:,i] = xv
            #intensity[i] = np.dot(bpv,soln[1:,i])

        return soln[1:,:]
                    

    def build_probe(self,__placements=None):
        '''
        build a default probe on the beginning and end of RNA
        '''
        self.updateparams()
        if __placements is None:
            __placements = np.zeros(self.N)
            __placements[3:10] = np.ones(7)
            #__placements[-10:-3] = np.ones(7)
            self.probe_design_vector = np.cumsum(__placements)
            #self.probe_design_vector = __placements
        else:
            self.probe_design_vector = np.cumsum(__placements)
        
    def build_W1(self):
        self.updateparams()
        __main_diag = [-1]*(self.N)
        W1=np.multiply(spdiags(np.array([__main_diag,[1]*self.N]),np.array([0,-1]),self.N,self.N),self.ke)
        return W1.toarray()

    def build_W0(self):
        self.updateparams()
        W0= np.atleast_2d(np.zeros(self.N)).T
        W0[0]=self.kb
        return W0

    def getfc(self):
        return np.dot(range(self.N),self.fi)
    
    # def map_to_fluorescence(self,soln=None):
    #     try:
    #         self.soln.shape
    #     except:
    #         print("no solution saved to memory, run self.solve()")
    #     if soln is None:
    #         self.fluorescence = np.dot(self.probe_design_vector,self.soln)
    #     else:
    #         self.fluorescence = np.dot(self.probe_design_vector,soln)
    #     return self.fluorescence 
        
    def get_flu_acc(self):
        fdata=self.map_to_fluorescence()
        return self.get_xcc(fdata)

    def getxi(self):
        self.updateparams()
        return np.zeros(self.N)

class TranslateVars(TranslateODE):    
    def __init__(self):
        
        GenericODE.__init__(self)
        TranslateODE.__init__(self)
        self.ptimes=100
        self.ti = 0; self.tf =10.0;

        self.N=100        
        self.kb=.33
        self.ke = np.ones(100)
        self.kb = .033
        self.fi = .1
        

        self.S = self.getselongation()
        self.W0 = self.getW0()
        self.W1 = self.getW1()
        self.ODE = self.variance_ODE
        self.xi=np.zeros(self.N+self.N**2)
    
    def vsolve(self,steady_state=False):  
        '''
        wrap solve function to have changing parameters
        if steady_state is true, solve the covariances by solving 
        the Lyapunov function. 
        '''
        self.S = self.getselongation()
        self.W0 = self.getW0()
        self.W1 = self.getW1()
        self.ODE = self.variance_ODE
        self.xi=np.zeros(self.N+self.N**2)
        #self.build_probe()
        self._solve()
        
    def variance_ODE(self,x,t):
        '''
        build variance/covariance ODEs.
        '''
        # dSIG = S*W1*SIG + SIG*W1'*S' + S*diag((W1*MU + W0)')*S'         
        __MU = x[:self.N]
        __SIG = np.reshape(x[self.N:],(self.N,self.N))
        # Build diagonal matrix
        inner = np.dot(self.W1,np.array([__MU]).T)+self.W0
        self.diag_mat = np.diag(np.ravel(inner),k=0)
        # Compute RHS
        RHS_vars = np.ravel(np.dot(np.dot(self.S,self.W1),__SIG) + np.dot(np.dot(__SIG,self.W1.T),self.S.T) + np.dot(np.dot(self.S,self.diag_mat),self.S.T))
        RHS_means = np.ravel(np.dot(self.S,(np.dot(self.W1,np.atleast_2d(__MU).T)))+np.dot(self.S,self.W0))
        return np.concatenate((RHS_means,RHS_vars))    

class TranslateCorrs(TranslateVars):
    '''
    get auto and cross correlations
    '''    
    def __init__(self):
        GenericODE.__init__(self)
     
        #### Default Parameters ####
#        self.pnum=100
#        self.ti = 0; self.tf =10.0;
#        self.N = 100; self.ke = 5  
#        self.kb = 5
#        self.params = {}
#        self.params['ke'] = self.ke 
#        self.params['kb'] = self.kb
#        self.binary = None
#        # Get fundamental matrices.        
#        self.S = self.getselongation()
#        self.W0 = self.getW0()
#        self.W1 = self.getW1()
#        self.mu_ss = self.get_mean_SS()
#        self.var_ss = self.get_var_SS()
#        
#        # solve for covariance steady state using ode
#        self.xi= np.ravel(self.var_ss)
#        self.ODE = self.correlation_ODE

    def get_mean_SS(self):       
        '''
        get the steady state mean values for the system
        '''
        M1 = -1*np.linalg.inv(np.dot(self.S,self.W1))
        return np.dot(M1,np.dot(self.S,self.W0))
   
    def get_var_SS(self):
        '''
        get the steady state variance for the system
        by solving the lyapunov fn. 
        '''
        A = np.dot(self.S,self.W1)
        inner = np.dot(self.W1,self.mu_ss)+self.W0
        diag_mat = np.diag(np.ravel(inner),k=0)
        Q = np.dot(np.dot(self.S,diag_mat),self.S.T)
        return solve_lyapunov(A,-Q)
         
    def csolve(self):
        '''
        update parameters and solve correlations
        '''
        self.get_autonomous_matrix()
        # Get steady state covariance as IC. 
        #self.vsolve() 
        #self.xi= np.ravel(self.soln[self.N:,-1])
        self.mu_ss = self.get_mean_SS()
        self.var_ss = self.get_var_SS()
        #self.vsolve()
        self.xi= np.ravel(self.var_ss)
        self.ODE = self.correlation_ODE
#        self._solve()
        self.faster_solve()

    def get_autonomous_matrix(self):
        '''
        Gets the phi matrix
        '''
        # Update fundamental matrices if parameters have been changed
        self.S = self.getselongation()
        self.W0 = self.getW0()
        self.W1 = self.getW1()
        self.phi = np.dot(self.W1.T,self.S.T) 
        self.phi = -np.diag(self.ke)+np.diag(self.ke,-1)[:-1,:-1]

    def correlation_ODE(self,x,t):
        '''
        Build variance/covariance ODEs. 
        '''
        _tmp = np.dot(self.phi,x.reshape(self.N,self.N)) 
        return np.ravel(_tmp.reshape(self.N**2,1) )
    
    def full_solve(self):
        phi = sp.sparse.csc_matrix(self.phi)
        cc_phi = sp.sparce.csc_matrix(np.zeros((self.N**2,self.N)) )
    

    def faster_solve(self):
        '''
        a faster way to solve the system of ODEs using 
        expokit.
        '''
        phi = sp.sparse.csc_matrix(self.phi)
        self.tvec = self.gettvec()
        
        #self.soln = np.zeros((self.N,self.N,len(self.tvec)))
        
        self.soln = sp.sparse.csc_matrix( (self.N*self.N,len(self.tvec)) , dtype=float )
        
        xi = self.xi.reshape(self.N,self.N)
  
        xtmp = np.copy(xi)
        self.intensity = np.zeros(len(self.tvec))
 
        self.intensity[0] = self.map_to_fluorescence(xi)
        #print(self.phi)
        st = time.time()
        xv = np.array([xtmp[:,0]]).T

 
        st = time.time()
        for i in range(1,len(self.tvec)):
            for j in range(self.N):
                xv = xtmp[:,j] 
                try:
                    xv,m,v = expv(self.tvec[i]-self.tvec[i-1],phi,xv,tol = 1e-6,m=30)
                    
                    
                except:
                    xv = np.zeros(xv.shape)
                #xv[np.abs(xv)<1e-8]=0
                xtmp[:,j] = xv

            self.intensity[i] = self.map_to_fluorescence(xtmp)
            if self.intensity[i]<1e-6:
                break
                self.soln[:,j,i] = xv 
        #self.soln =self.soln.reshape(self.N**2,len(self.tvec))
    
        
        return self.intensity 

class TranslateCoarseCorrs(TranslateCorrs):
    '''
    get auto and cross correlations for Coarse Grained model
    '''    
    def __init__(self,model,N):
        self.N=N
        self.binsize=int(np.ceil(model.N/float(N)))
        self.loadfullmodel(model)
        
    def loadprobe(self,probedesign):
        self.Nreal=len(probedesign)
        binary=np.zeros(self.N)
        for i in range(self.N):
            binary[i]=np.sum(probedesign[i*self.binsize:(i+1)*self.binsize])
        return binary
        
    def loadfullmodel(self,model):
        GenericODE.__init__(self)
        self.ptimes=model.ptimes
        self.ti = model.ti
        self.tf = model.tf
        self.Nreal = model.N 
        self.ke = model.ke/self.N
        self.kb = model.kb
        self.fi = model.fi
        self.params = {}
        for key,val in model.params.iteritems():
            if key is 'ke':
                self.params['ke'] = float(model.ke)*self.binsize
            else:
                self.params[key] = val
        self.binary = self.loadprobe(model.binary)
        self.build_probe(self.binary)
        
        
       
        # Get fundamental matrices.        
        self.S = self.getselongation()
        self.W0 = self.getW0()
        self.W1 = self.getW1()
        self.mu_ss = self.get_mean_SS()
        self.var_ss = self.get_var_SS()
        
        # solve for covariance steady state using ode
        self.xi= np.ravel(self.var_ss)
        self.ODE = self.correlation_ODE

        #_tmp = np.dot(np.dot(self.W1.T,self.S.T),x.reshape(self.N,self.N))
        #return np.ravel(_tmp.reshape(self.N**2,1) )
       # Ci = np.reshape(cc.soln[:,i],(cc.N,cc.N))



# -*- coding: utf-8 -*-
"""
Created on Tue Apr 21 16:29:21 2020

@author: William Raymond
"""

import os
import re

from snapgene_reader import snapgene_file_to_seqrecord
from Bio import SeqIO


class Error(Exception):
    """Base class for exceptions in this module."""
    pass

class SnapGeneMissingError(Error):
    """Exception raised for errors in the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

class FileTypeNotRecognizedError(Error):
    """Exception raised for errors in the input.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message

class FileParser():
    '''
    Class to parse incoming files

    TODO: Extend this for psuedouridine /m6a / modification dictionary
    '''
    def __init__(self):
        self.file = None
        self.letter_dict = ['a', 'c', 'u', 'g']
        # IPUAC SUBSTITIONS - They need to  adjusted /
        #extended for mrna modifications?
        self.sub_dict = {'m':'a', 'r':'a', 'y':'t', 'k':'g', 's':'g',
                         'w':'a', 'h':'a', 'n':'a', 'Ψ':'u'}

        self.test_seq = 'aucuguacguacguaucgaucguguacuggcaaaacgu'\
                        'aguagcugagcaucaucuaug'


    def clean_seq(self, seq):
        '''
        Return an mrna sequence of lowercase a,u,c,g from IPUAC substitutions

        .. warning:: this code will replace substitutive nucleotides with
        preferential order a, g , u , c. for example: N (any base) is set
        to A, W (T,U, or A) is set to A, S (C or G) is set to G



        Parameters
        ----------
        seq : str
            sequence string.

        Returns
        -------
        seq : str
            cleaned sequence str (only lowercase a,u,c,g).

        '''
        seq = seq.lower()

        for key in self.sub_dict.keys():
            seq = seq.replace(key, self.sub_dict[key])

        seq = seq.replace('t', 'u')
        return seq


    def get_sequence(self, file):
        '''
        get a nucleotide sequence

        Parameters
        ----------
        file : path
            Path to the file to open.

        Returns
        -------
        str
            mRNA sequence string.

        '''
        self.__check_valid_file(file)
        extension = file.split('.')[-1]
        if extension == 'dna':
            try:
                seq_record = snapgene_file_to_seqrecord(file)
            except:
                msg = 'To read .dna files please install snapegenereader: '\
                      ' pip install snapgene_reader - '\
                          'https://github.com/IsaacLuo/SnapGeneFileReader'
                raise SnapGeneMissingError(msg)
            sequence_str = str(seq_record.seq)

        if extension == 'txt':
            sequence_str = self.__get_seq_from_txt(file)


        if extension == 'gb':
            gb_record = SeqIO.read(open(file, "r"), "genbank")
            sequence_str = str(gb_record.seq)

        if extension == 'fasta':
            fastas = list(SeqIO.parse(file, 'fasta'))
            if len(fastas) > 1:
                return 'Multiple line fastas not supported'
            else:
                sequence_str = (str(fastas[0].seq))
        cleaned_sequence_str = self.clean_seq(sequence_str)

        return cleaned_sequence_str

    def __check_valid_file(self, file):
        extension = file.split('.')[-1]

        if extension in ['fasta', 'gb', 'txt', 'dna']:
            return True
        else:
            raise FileTypeNotRecognizedError("Unrecognized File '\
                                             'type, the sequence '\
                            'file must be a .txt, .dna, .gb, or .fasta")



    def get_name(self, file_path):
        '''
        attempt to find the transcript name from a file

        Parameters
        ----------
        file_path : str
            string path to file.

        Returns
        -------
        str
            transcript name or "unknown".

        '''
        self.__check_valid_file(file_path)

        name = 'unknown'
        extension = file_path.split('.')[-1]
        if extension == 'fasta':
            fastas = list(SeqIO.parse(file_path, 'fasta'))
            if len(fastas) > 1:
                return 'Multiple line fastas not supported'
            else:
                name = (str(fastas[0].id))

        if extension == 'gb':
            gb_record = SeqIO.read(open(file_path, "r"), "genbank")
            name = str(gb_record.id)

        if extension == 'txt':
            name = self.__get_name_from_text(file_path)

        if extension == 'dna':
            try:
                seq_record = snapgene_file_to_seqrecord(file_path)
            except:
                msg = 'To read .dna files please install snapegenereader: '\
                      ' pip install snapgene_reader - '\
                          'https://github.com/IsaacLuo/SnapGeneFileReader'
                raise SnapGeneMissingError(msg)
            name = seq_record.name

        return name


    def get_description(self, file_path):
        '''
        Attempt to find the text description from a file

        Parameters
        ----------
        file_path : str
            string path to file.

        Returns
        -------
        str
            transcript description or "unknown".

        '''
        self.__check_valid_file(file_path)

        extension = file_path.split('.')[-1]

        if extension == 'fasta':
            fastas = list(SeqIO.parse(file_path, 'fasta'))
            if len(fastas) > 1:
                return 'Multiple line fastas not supported'
            else:

                desc = str(fastas[0].description)

        if extension == 'gb':
            gb_record = SeqIO.read(open(file_path, "r"), "genbank")
            desc = str(gb_record.description)

        if extension == 'dna':
            try:
                seq_record = snapgene_file_to_seqrecord(file_path)
            except:
                msg = 'To read .dna files please install snapegenereader: '\
                      ' pip install snapgene_reader - '\
                          'https://github.com/IsaacLuo/SnapGeneFileReader'
                raise SnapGeneMissingError(msg)

            desc = seq_record.description
        if extension == 'txt':
            desc = '<unknown description>'

        return desc



    @classmethod
    def __get_seq_from_txt(cls, file):
        with open(file) as fname:
            raw = fname.readlines()

        raw = ''.join(raw)
        onlychar = re.split(r'[^A-Za-z]', raw)
        validt = ['A', 'G', 'T', 'C']
        validu = ['A', 'G', 'U', 'C']

        sequence_str = ''

        for i in range(len(onlychar)):
            section = onlychar[i]
            if set(section.upper()) == set(validt):
                sequence_str += section.upper()


            elif set(section.upper()) == set(validu):
                sequence_str += section.upper()
        return sequence_str

    @classmethod
    def __get_name_from_text(cls, file):
        name = ''
        with open(file) as fname:
            raw = fname.readlines()
            raw = ''.join(raw)

            validt = ['A', 'G', 'T', 'C']

            validu = ['A', 'G', 'U', 'C']
            onlychar = re.split(r'[^A-Za-z]', raw)

            namelen = 0

            for i in range(len(onlychar)):
                section = onlychar[i]
                if set(section.upper()) != set(validt):
                    if set(section.upper()) != set(validu):
                        if len(section) > namelen:
                            name = section
                            namelen = len(section)

        if name == '':
            name = os.path.basename(file)[:-4]

        return name



# -*- coding: utf-8 -*-
"""
Created on Thu May 21 16:02:42 2020

@author: willi
"""

import numpy as np
import ssa_translation_lowmem
import ssa_translation_lowmem_leaky
import ssa_translation_lowmem_nostats
import ssa_translation
import matplotlib.pyplot as plt
import time
import os

os.chdir('..')
from rss import ProbeVectorFactory as pvf
from rss import PropensityFactory as pff
os.chdir('ssa_cpp')


# load the elongation 
kelong = np.loadtxt('elongationrates.txt')
kbind = kelong[0]
kcompl = kelong[-1]
kelong = kelong[1:-1]
ncolor = 2
t_array = np.array([0,10,20,30,50,100,250,500],dtype=np.float64)
t0 = 15
t_array = np.linspace(0,1000,1000,dtype=np.float64)
N_rib = 200
result = np.zeros((len(t_array)*ncolor),dtype=np.int32  )
#kelong = np.array([3.1,3.2,3.3,3.4,3.5,3.1,3.2,3.3,3.4,3.5],dtype=np.float64)
n_trajectories = 100


#preallocated arrays here
all_results = np.zeros((n_trajectories,len(t_array),ncolor),dtype=np.int32)
lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t_array<20)[0]))

all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)

all_ribtimes = np.zeros((n_trajectories,400),dtype=np.float64)
all_coltimes = np.zeros((n_trajectories,400),dtype=np.int32)
nribs = np.array([0],dtype=np.int32)
all_ribs = np.zeros((n_trajectories,1))
seeds = np.random.randint(0,0x7FFFFFF,n_trajectories)
x0 = np.zeros((N_rib),dtype=np.int32)

pl = np.zeros((len(kelong),ncolor), dtype=np.int32)

pl[ [10,20,30,100,120,140],0  ] = 1
#pl[ [10,140],1  ] = 1

pl = np.cumsum(pl,axis=0)
pl = pl.T.copy(order='C')



print('-----------------------')
print('GENERATING REPORT')
print('-----------------------')
print('1 cpu core, 2 color')
print('{0} base pairs'.format(len(kelong)-2))
print('-----------------------')



all_col_points = []
start = time.time()
for i in range(n_trajectories):
    result = np.zeros((ncolor,len(t_array)),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    
    colpointsx = np.zeros(len(kelong)*400,dtype=np.int32)
    colpointst = np.zeros(len(kelong)*400,dtype=np.float64)
    
    ssa_translation_lowmem.run_SSA(result,ribtimes,coltimes,colpointsx,colpointst, kelong,frapresult,t_array,.03,kcompl, 1,0,300, seeds[i],nribs,x0,9, pl,2)
    

  
    all_results[i,:,:] = result.T
    
    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    endcolrec = np.where(colpointsx == 0)[0][0]
    
    colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
    all_col_points.append(colpoints.T)
print('low memory w/recording_stats: time for {0} trajectories {1}'.format(n_trajectories,time.time()-start))
#plt.hist(result[result>0])
#plt.show()
#traj = result.reshape((N_rib,len(t_array))).T
##print('The result is \n {0}'.format(result.reshape((N_rib,len(t_array))).T))
#plt.plot(traj[-1,:])
#plt.show()

plt.plot(result.T,'--')

start = time.time()
for i in range(n_trajectories):
    result = np.zeros((ncolor,len(t_array)),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    
    ssa_translation_lowmem_nostats.run_SSA(result, kelong,frapresult,t_array,.03,kcompl, 1,0,300, seeds[i],x0,9, pl,2)
    

  
    all_results[i,:,:] = result.T
    
    all_frapresults[i,:] = frapresult


print('Low memory w/o recording stats: time for {0} trajectories {1}'.format(n_trajectories,time.time()-start))




plt.plot(result.T)




pl_2color = np.atleast_2d(pl)
probe_loc = (pl_2color[:,1:]-pl_2color[:,:-1] > 0).astype(int)
inds = pff.intellegent_bin(np.atleast_2d(probe_loc),100)
bpv,bpl = pvf.bin_probe_vecs(probe_loc,inds)

kelong_2color = kelong
k_bin = pff.bin_k(kelong, inds)




start = time.time()
for i in range(n_trajectories):
    result = np.zeros((ncolor,len(t_array)),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    
    ssa_translation_lowmem_nostats.run_SSA(result, k_bin,frapresult,t_array,.03,kcompl, 1,0,300, seeds[i],x0,9, bpl,2)
    

  
    #all_results[i,:,:] = result.T
    
    #all_frapresults[i,:] = frapresult


print('Low memory 100 bins: time for {0} trajectories {1}'.format(n_trajectories,time.time()-start))



plt.plot(result.T,'.')



plt.legend(['w/ stats color 1','w/ stats color 2','w/o stats color 1','w/o stats color 2'])
plt.xlabel('time')
plt.ylabel('intensity')


print('-----------------------')
print('1 cpu core, 1 color')
print('-----------------------')

all_results = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)
lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t_array<20)[0]))

all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)

all_ribtimes = np.zeros((n_trajectories,400),dtype=np.float64)
all_coltimes = np.zeros((n_trajectories,400),dtype=np.int32)
nribs = np.array([0],dtype=np.int32)
all_ribs = np.zeros((n_trajectories,1))
seeds = np.random.randint(0,0x7FFFFFF,n_trajectories)
x0 = np.zeros((N_rib),dtype=np.int32)

pv = np.loadtxt('probe_design.txt')
all_col_points = []
start = time.time()



for i in range(n_trajectories):
    result = np.zeros((len(t_array)*N_rib),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    
    colpointsx = np.zeros(len(kelong)*400,dtype=np.int32)
    colpointst = np.zeros(len(kelong)*400,dtype=np.float64)
    
    
    
    print(result.shape)
    print(kelong.shape)
    print(frapresult.shape)
    print(ribtimes.shape)
    print(coltimes.shape)
    print(colpointsx.shape)
    print(colpointst.shape)
    
    
    ssa_translation.run_SSA(result,ribtimes,coltimes,colpointsx,colpointst, kelong,frapresult,t_array,.03,kcompl, 1,0,300, seeds[i],nribs,x0,9,N_rib)
    
    

    all_results[i,:] = result

    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    endcolrec = np.where(colpointsx == 0)[0][0]
    
    colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
    all_col_points.append(colpoints.T)
    
ntimes = len(t_array)
intensity_vec = np.zeros(ntimes)

tstart = 0
I = np.zeros((n_trajectories,ntimes-tstart))
for i in range(n_trajectories):
    traj = all_results[i,:].reshape((N_rib,len(t_array))).T
    for j in range(tstart,ntimes):
        temp_output = traj[j,:]
        I[i,j] = np.sum(pv[temp_output[temp_output>0]-1])
        
print('Full SSA with Recording: time for {0} trajectories {1}'.format(n_trajectories,time.time()-start))
plt.figure()
plt.plot(I[-1,:],'x')



pl = np.atleast_2d(pv.astype(int))
ncolor=1

all_results = np.zeros((n_trajectories,len(t_array),ncolor),dtype=np.int32)
lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t_array<20)[0]))

all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)

all_ribtimes = np.zeros((n_trajectories,400),dtype=np.float64)
all_coltimes = np.zeros((n_trajectories,400),dtype=np.int32)
nribs = np.array([0],dtype=np.int32)
all_ribs = np.zeros((n_trajectories,1))
#seeds = np.random.randint(0,0x7FFFFFF,n_trajectories)
x0 = np.zeros((N_rib),dtype=np.int32)


all_col_points = []
start = time.time()
for i in range(n_trajectories):
    result = np.zeros((ncolor,len(t_array)),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    
    colpointsx = np.zeros(len(kelong)*400,dtype=np.int32)
    colpointst = np.zeros(len(kelong)*400,dtype=np.float64)
    
    ssa_translation_lowmem.run_SSA(result,ribtimes,coltimes,colpointsx,colpointst, kelong,frapresult,t_array,.03,kcompl, 1,0,300, seeds[i],nribs,x0,9, pl,1)
    

  
    all_results[i,:] = result.T
    
    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    endcolrec = np.where(colpointsx == 0)[0][0]
    
    colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
    all_col_points.append(colpoints.T)
print('Low memory w/recording_stats: time for {0} trajectories {1}'.format(n_trajectories,time.time()-start))
#plt.hist(result[result>0])
#plt.show()
#traj = result.reshape((N_rib,len(t_array))).T
##print('The result is \n {0}'.format(result.reshape((N_rib,len(t_array))).T))
#plt.plot(traj[-1,:])
#plt.show()
plt.plot(result.T,'o')


start = time.time()
for i in range(n_trajectories):
    result = np.zeros((ncolor,len(t_array)),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    
    ssa_translation_lowmem_nostats.run_SSA(result, kelong,frapresult,t_array,.03,kcompl, 1,0,300, seeds[i],x0,9, pl,1)
    

  
    all_results[i,:] = result.T
    
    all_frapresults[i,:] = frapresult


print('Low memory w/o recording stats: time for {0} trajectories {1}'.format(n_trajectories,time.time()-start))


plt.plot(result.T)

pl = pl.flatten()
probe_loc = (np.where(pv[1:]-pv[:-1] > 0)[0]+1).astype(np.int32)
k_probe = .2

all_results = np.zeros((n_trajectories,len(t_array)),dtype=np.int32)
lenfrap = len(np.intersect1d(np.where(t_array>0)[0],np.where(t_array<20)[0]))

all_frapresults = np.zeros((n_trajectories,N_rib*len(t_array)),dtype=np.int32)

all_ribtimes = np.zeros((n_trajectories,400),dtype=np.float64)
all_coltimes = np.zeros((n_trajectories,400),dtype=np.int32)
nribs = np.array([0],dtype=np.int32)
all_ribs = np.zeros((n_trajectories,1))
seeds = np.random.randint(0,0x7FFFFFF,n_trajectories)
x0 = np.zeros((N_rib),dtype=np.int32)

start = time.time()
for i in range(n_trajectories):
    result = np.zeros((len(t_array)),dtype=np.int32)    
    frapresult = np.zeros((len(t_array)*N_rib),dtype=np.int32)
    
    ribtimes = np.zeros((400),dtype=np.float64)
    coltimes = np.zeros((400),dtype=np.int32)
    
    colpointsx = np.zeros(len(kelong)*400,dtype=np.int32)
    colpointst = np.zeros(len(kelong)*400,dtype=np.float64)
    
    ssa_translation_lowmem_leaky.run_SSA(result,ribtimes,coltimes,colpointsx,colpointst, kelong,frapresult,t_array,.03,kcompl, 1,0,300, seeds[i],nribs,x0,9, pl,k_probe,probe_loc )
    
  
    all_results[i,:] = result
    
    all_frapresults[i,:] = frapresult
    all_coltimes[i,:] = coltimes
    all_ribtimes[i,:] = ribtimes
    all_ribs[i,:] = nribs[0]

    endcolrec = np.where(colpointsx == 0)[0][0]
    
    colpoints = np.vstack((colpointsx[:endcolrec],colpointst[:endcolrec]))
    all_col_points.append(colpoints.T)
print('Low memory w/ leaky probes: time for {0} trajectories {1}'.format(n_trajectories,time.time()-start))

plt.plot(result.T)
plt.legend(['full ssa','lowmem w/ stats','lowmem w/o stats','lowmem leaky'])
plt.xlabel('time')
plt.ylabel('intensity')


# -*- coding: utf-8 -*-
"""
Created on Fri Aug 20 11:21:05 2021

@author: willi
"""

import time
import numpy as np
import os
import importlib
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib import cm
import matplotlib.patches as mpatches

from . import rsnapsim_model_maker
ModelFactory = rsnapsim_model_maker.ModelFactory
RuleConverter = rsnapsim_model_maker.RuleConverter
from . import PropensityFactory, ProbeVectorFactory


class ModelBuilder():
    def __init__(self):
        self.poi = None
        self.n_total_reactions = 0
        self.n_state_reactions = 0
        self.max_ribs = 100
        self.parameters = None
        self.rules = None
        self.kin = 0.033
        self.kout = 10
        
    
    def get_max_ribs(self):
        self.max_rib = int(len(self.poi.kelong)/9 + 5)
        
    def build_lattice(self, N_reactions):
        forward_rates = self.poi.kelong
        self.stoich_lattice = np.zeros([N_reactions,
                                        len(forward_rates)], dtype=np.int32)
        self.xi_lattice = np.zeros([1,len(forward_rates)], dtype=np.int32)
        self.n_total_reactions += N_reactions
        
    def build_states(self, state_stoichiometry, state_inital_condition):
        self.stoich_state = state_stoichiometry.astype(np.int32)
        self.xi_state = state_inital_condition.astype(np.int32)
        
        self.n_total_reactions += int(state_stoichiometry.shape[0])
        self.n_state_reactions = int(state_stoichiometry.shape[0])
        
    
    def compile_model(self, name, overwrite=True, rules = None, eigen_path=''):
        mf = ModelFactory()
        mf.compile_model(name, overwrite=True, rules = rules, eigen_path=eigen_path)
        model = self.get_model(name)
        
    def get_model(self, name):
        model = importlib.import_module('rsnapsim.models.%s.%s'%(name,name))
        return model
    
    

            
                
                
        
               
        
        


# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 17:56:11 2020

@author: willi
"""


from . import GenericMetaData
GenericMetaData = GenericMetaData.GenericMetaData
import numpy as np
import json, codecs

class ODE_Soln():
    def __init__(self):
        
        self.__arrays = [    'probe_loc',
                             'x0',
                             'k',
                             'N',
                             'time',
                             'mu_It',
                             'intensity_acc',
                             'intensity_acc_norm',]
        
        self.__arrays_2d = ['var_state_ss']
        
        self.__vals = ['fi','kb','N','solve_time','mu_I_ss','var_I_ss']
        self.__meta = GenericMetaData().get()
        
        
        pass
    
    def save(self,filename):
        ext = filename.split('.')[-1]
        
        if 'txt' == ext:
            self.__save_txt(filename)
        if 'json' == ext:
            self.__save_json(filename)
            

    def load(self,filename):
        ext = filename.split('.')[-1]
        
        if 'txt' == ext:
            self.__load_from_txt(filename)
        if 'json' == ext:
            self.__load_from_json(filename)
            
    def __save_txt(self,filename):

        if '.txt' not in filename:    
            filename = filename + '.txt'


      
        f = open(filename,'w')
        for key in self.__dict__.keys():
            
            if key in self.__arrays + self.__arrays_2d + self.__vals:

                f.write((key + '\r\n'))
                np.savetxt(f, np.atleast_2d(self.__dict__[key]), delimiter=',', fmt='%s')
                f.write(('\r\n'))
                
        f.close()

    def __load_from_txt(self, filename):
        if '.txt' in filename:
            ode_obj = np.loadtxt(filename, dtype=str,delimiter='\n')
          
            solutions = []
            for i in range(0,len(ode_obj)-1):
                label = ode_obj[i]                
                
                if label in self.__arrays:
                    
                    array = np.fromstring(ode_obj[i+1], dtype=float, sep=',')
                    exec(('self.'+label+ '=array'))  
                
                if label in self.__vals:
                    value = np.fromstring(ode_obj[i+1], dtype=float, sep=',')
                    exec(('self.'+label+'=value'))
                    
                if label in self.__arrays_2d:
                    j = 1
                    array2d = np.array([[]])
                    while ode_obj[i+j] not in self.__arrays + self.__vals:
                        if j == 1:
                            array2d = np.array([np.fromstring(ode_obj[i+j], dtype=float, sep=',')   ])
                        else:
                            str_2d = np.fromstring(ode_obj[i+j], dtype=float, sep=',')    
                            array2d = np.vstack((array2d,str_2d))
                        
                        j+=1
                        
                   
                    exec(('self.'+label+'=array2d'))
                        
                


    def __save_from_json(self, filename):

        if '.json' not in filename:
            filename =  filename + '.json'

        odedict = {}
        for key in self.__dict__.keys():
            if key in self.__arrays:
                
                odedict[key] = self.ssa_harr.__dict__[key].tolist()
            else:
                odedict[key] = self.ssa_harr.__dict__[key]

        json.dump(odedict, codecs.open(filename, 'w', encoding='utf-8'), separators=(',', ':'), sort_keys=True, indent=4)


    def __load_json(self,filename):
        if '.json' in filename:

            obj_text = codecs.open(filename, 'r', encoding='utf-8').read()
            ode_dict = json.loads(obj_text)

            for key in ode_dict.keys():
                if key in self.__arrays:

                    self.__dict__[key] = np.array(ode_dict[key])
                else:
                    self.__dict__[key] = ode_dict[key]


# -*- coding: utf-8 -*-
"""
Created on Thu Dec 17 18:06:56 2020

@author: William Raymond
"""
import numpy as np


'''
This is a dictionary object that handles everything dealing with tag sequences and AA tables
'''


class CodonDictionaries():
    '''
    Attributes
    ----------

    tag_dict: dict
        dictionary of tag name and their epitopes
    tag_colors: dict
        dictionary of tag name and their colors
    tag_full: dict
        dictionary of tag name and their full nt sequence
    aa_keys: list
        keys for amino acids
    aa_table: dict
        codon to aa table
    aa_table_r: dict
        aa to codons table
    human_codon_frequency_bias_nakamura: dict
        codon gene copys for both U|T
    human_codon_frequency_bias_nakamura_single: dict
        a codon gene copy number with only T, for average /
        calculation basis that would be thrown off with including U and T
    trna_ids: list
        tRNA species to ID

    '''


    def __init__(self):
        self.tag_dict = {'T_SunTag':'EELLSKNYHLENEVARLKK',
                         'T_Flag':'DYKDDDDK',
                         'T_Hemagglutinin':'YPYDVPDYA'}

        self.tag_colors = {'T_SunTag':'green',
                           'T_Flag':'blue',
                           'T_Hemagglutinin':'blue'}

        self.tag_full = {'T_Flag':('ATGGACTACAAGGACGACGACGACAAAGGTGAC'
                                   'TACAAAGATGATGACGATAAAGGCGACTATA'
                                   'AGGACGATGACGACAAGGGCGGAAACTCACTGA'
                                   'TCAAGGAAAACATGCGGATGAAGGTGGTGAT'
                                   'GGAGGGCTCCGTGAATGGTCACCAGTTCAAGTG'
                                   'CACCGGAGAGGGAGAGGGAAACCCGTACATG'
                                   'GGAACTCAGACCATGCGCATTAAGGTCATCGAA'
                                   'GGAGGTCCGCTGCCGTTCGCTTTCGATATCC'
                                   'TGGCCACTTCGTTCGGAGGAGGGTCGCGCACGTTC'
                                   'ATCAAGTACCCGAAGGGAATCCCGGACTT'
                                   'CTTTAAGCAGTCATTCCCGGAAGGATTCACTTGGG'
                                   'AACGGGTGACCCGGTATGAAGATGGAGGT'
                                   'GTGGTGACTGTCATGCAAGATACTTCGCTGGAGGATGGG'
                                   'TGCCTCGTGTACCACGTCCAAGTCC'
                                   'GCGGAGTGAATTTCCCGTCCAACGGACCAGTGATGCAG'
                                   'AAAAAGACGAAGGGTTGGGAACCTAA'
                                   'TACTGAAATGATGTACCCCGCAGACGGAGGGCTGAGGG'
                                   'GCTACACCCACATGGCGCTGAAGGTC'
                                   'GACGGAGGAGATTACAAGGATGACGACGATAAGCAACAA'
                                   'GATTACAAAGACGATGATGACAAGG'
                                   'GCCAGCAGGGCGACTACAAGGACGACGACGACAAGCAG'
                                   'CAGGACTACAAAGATGACGATGATAA'
                                   'AGGAGGAGGACATCTGTCCTGTTCGTTCGTGACCACCT'
                                   'ACAGATCAAAGAAAACCGTGGGAAAC'
                                   'ATCAAGATGCCGGGCATTCATGCCGTCGACCACCGCCT'
                                   'GGAGCGGCTCGAAGAATCAGACAATG'
                                   'AGATGTTCGTCGTGCAAAGAGAACATGCCGTGGCCAAGTT'
                                   'CGCGGGACTGGGAGGCGGTGGAGG'
                                   'CGATTACAAAGACGATGATGACAAGGGTGACTATAAAGA'
                                   'CGACGATGACAAAGGGGATTACAAG'
                                   'GATGATGATGATAAGGGAGGCGGTGGATCAGGTGGAG'
                                   'GAGGTTCACTGCAG')}

        self.aa_keys = ['A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G', 'H', 'I',
                        'L', 'K', 'M', 'F',
                        'P', 'S', 'T', 'W', 'Y', 'V', '*']

        self.codon_types = dict(zip(
            self.aa_keys, np.ones((1, 21)).flatten().astype(int).tolist()))

        self.aa_table = {
            'ATA':'I', 'ATC':'I', 'ATT':'I', 'ATG':'M',
            'ACA':'T', 'ACC':'T', 'ACG':'T', 'ACT':'T',
            'AAC':'N', 'AAT':'N', 'AAA':'K', 'AAG':'K',
            'AGC':'S', 'AGT':'S', 'AGA':'R', 'AGG':'R',
            'CTA':'L', 'CTC':'L', 'CTG':'L', 'CTT':'L',
            'CCA':'P', 'CCC':'P', 'CCG':'P', 'CCT':'P',
            'CAC':'H', 'CAT':'H', 'CAA':'Q', 'CAG':'Q',
            'CGA':'R', 'CGC':'R', 'CGG':'R', 'CGT':'R',
            'GTA':'V', 'GTC':'V', 'GTG':'V', 'GTT':'V',
            'GCA':'A', 'GCC':'A', 'GCG':'A', 'GCT':'A',
            'GAC':'D', 'GAT':'D', 'GAA':'E', 'GAG':'E',
            'GGA':'G', 'GGC':'G', 'GGG':'G', 'GGT':'G',
            'TCA':'S', 'TCC':'S', 'TCG':'S', 'TCT':'S',
            'TTC':'F', 'TTT':'F', 'TTA':'L', 'TTG':'L',
            'TAC':'Y', 'TAT':'Y', 'TAA':'*', 'TAG':'*',
            'TGC':'C', 'TGT':'C', 'TGA':'*', 'TGG':'W',

            'AUA':'I', 'AUC':'I', 'AUU':'I', 'AUG':'M',
            'ACU':'T',
            'AAU':'N',
            'AGU':'S',
            'CUA':'L', 'CUC':'L', 'CUG':'L', 'CUU':'L',
            'CCU':'P',
            'CAU':'H',
            'CGU':'R',
            'GUA':'V', 'GUC':'V', 'GUG':'V', 'GUU':'V',
            'GCU':'A',
            'GAU':'D',
            'GGU':'G',
            'UCA':'S', 'UCC':'S', 'UCG':'S', 'UCU':'S',
            'UUC':'F', 'UUU':'F', 'UUA':'L', 'UUG':'L',
            'UAC':'Y', 'UAU':'Y', 'UAA':'*', 'UAG':'*',
            'UGC':'C', 'UGU':'C', 'UGA':'*', 'UGG':'W',}

        self.aa_table_r = {
            'A':['GCA', 'GCC', 'GCG', 'GCT', 'GCU'],
            'R':['CGA', 'CGC', 'CGG', 'CGT', 'AGG', 'AGA', 'CGU'],
            'N':['AAC', 'AAT', 'AAU'],
            'D':['GAC', 'GAT', 'GAU'],
            'C':['TGC', 'TGT', 'UGC', 'UGU'],
            'Q':['CAA', 'CAG'],
            'E':['GAA', 'GAG'],
            'G':['GGT', 'GGC', 'GGA', 'GGG', 'GGU'],
            'H':['CAC', 'CAT', 'CAU'],
            'I':['ATT', 'ATC', 'ATA', 'AUU', 'AUC', 'AUA'],
            'L':['CTA', 'CTC', 'CTG', 'CTT', 'TTA', 'TTG', 'CUA',
                 'CUC', 'CUG', 'CUU', 'UUA', 'UUG'],
            'K':['AAA', 'AAG'],
            'M':['ATG', 'AUG'],
            'F':['TTC', 'TTT', 'UUC', 'UUU'],
            'P':['CCT', 'CCC', 'CCG', 'CCA', 'CCU'],
            'S':['TCA', 'TCC', 'TCG', 'TCT', 'AGC', 'AGT',
                 'UCA', 'UCC', 'UCG', 'UCU', 'AGU'],
            'T':['ACA', 'ACC', 'ACG', 'ACT', 'ACU'],
            'W':['TGG', 'UGG'],
            'Y':['TAT', 'TAC', 'UAC', 'UAU'],
            'V':['GTA', 'GTC', 'GTT', 'GTG', 'GUG', 'GUU',
                 'GUC', 'GUA'],
            '*':['TGA', 'TAG', 'TAA', 'UGA', 'UAG', 'UAA']
            }


        self.human_codon_frequency_bias_nakamura = {
            'TTT': 17.6, 'TCT': 15.2, 'TAT': 12.2, 'TGT': 10.6, 'TTC': 20.3,
            'TCC': 17.7, 'TAC': 15.3, 'TGC': 12.6, 'TTA': 7.7, 'TCA': 12.2,
            'TAA': 1.0, 'TGA': 1.6, 'TTG': 12.9, 'TCG':  4.4, 'TAG': 0.8,
            'TGG': 13.2, 'CTT': 13.2, 'CCT': 17.5, 'CAT': 10.9, 'CGT': 4.5,
            'CTC': 19.6, 'CCC': 19.8, 'CAC': 15.1, 'CGC': 10.4, 'CTA':  7.2,
            'CCA': 16.9, 'CAA': 12.3, 'CGA':  6.2, 'CTG': 39.6, 'CCG':  6.9,
            'CAG': 34.2, 'CGG': 11.4, 'ATT': 16.0, 'ACT': 13.1, 'AAT': 17.0,
            'AGT': 12.1, 'ATC': 20.8, 'ACC': 18.9, 'AAC': 19.1, 'AGC': 19.5,
            'ATA':  7.5, 'ACA': 15.1, 'AAA': 24.4, 'AGA': 12.2, 'ATG': 22.0,
            'ACG': 6.1, 'AAG': 31.9, 'AGG': 12.0, 'GTT': 11.0, 'GCT': 18.4,
            'GAT': 21.8, 'GGT': 10.8, 'GTC': 14.5, 'GCC': 27.7, 'GAC': 25.1,
            'GGC': 22.2, 'GTA':  7.1, 'GCA': 15.8, 'GAA': 29.0, 'GGA': 16.5,
            'GTG': 28.1, 'GCG': 7.4, 'GAG': 39.6, 'GGG': 16.5}

        self.human_codon_frequency_bias_nakamura_single = {
            'TTT': 17.6, 'TCT': 15.2, 'TAT': 12.2, 'TGT': 10.6, 'TTC': 20.3,
            'TCC': 17.7, 'TAC': 15.3, 'TGC': 12.6, 'TTA': 7.7, 'TCA': 12.2,
            'TAA': 1.0, 'TGA': 1.6, 'TTG': 12.9, 'TCG':  4.4, 'TAG': 0.8,
            'TGG': 13.2, 'CTT': 13.2, 'CCT': 17.5, 'CAT': 10.9, 'CGT': 4.5,
            'CTC': 19.6, 'CCC': 19.8, 'CAC': 15.1, 'CGC': 10.4, 'CTA':  7.2,
            'CCA': 16.9, 'CAA': 12.3, 'CGA':  6.2, 'CTG': 39.6, 'CCG':  6.9,
            'CAG': 34.2, 'CGG': 11.4, 'ATT': 16.0, 'ACT': 13.1, 'AAT': 17.0,
            'AGT': 12.1, 'ATC': 20.8, 'ACC': 18.9, 'AAC': 19.1, 'AGC': 19.5,
            'ATA':  7.5, 'ACA': 15.1, 'AAA': 24.4, 'AGA': 12.2, 'ATG': 22.0,
            'ACG': 6.1, 'AAG': 31.9, 'AGG': 12.0, 'GTT': 11.0, 'GCT': 18.4,
            'GAT': 21.8, 'GGT': 10.8, 'GTC': 14.5, 'GCC': 27.7, 'GAC': 25.1,
            'GGC': 22.2, 'GTA':  7.1, 'GCA': 15.8, 'GAA': 29.0, 'GGA': 16.5,
            'GTG': 28.1, 'GCG': 7.4, 'GAG': 39.6, 'GGG': 16.5}



        self.trna_ids = [
            'TTT', 'TCT', 'TAT', 'TGT', 'TTC', 'TCC',
            'TAC', 'TGC', 'TTA', 'TCA', 'TAA',
            'TGA', 'TTG', 'TCG', 'TAG', 'TGG', 'CTT', 'CCT', 'CAT',
            'CGT', 'CTC', 'CCC', 'CAC', 'CGC', 'CTA',
            'CCA', 'CAA', 'CGA', 'CTG', 'CCG', 'CAG', 'CGG', 'ATT',
            'ACT', 'AAT', 'AGT', 'ATC',
            'ACC', 'AAC', 'AGC', 'ATA', 'ACA', 'AAA', 'AGA', 'ATG',
            'ACG', 'AAG', 'AGG', 'GTT',
            'GCT', 'GAT', 'GGT', 'GTC', 'GCC', 'GAC', 'GGC', 'GTA',
            'GCA', 'GAA', 'GGA', 'GTG',
            'GCG', 'GAG', 'GGG']

        self.trna_ids.remove('TAG')
        self.trna_ids.remove('TAA')
        self.trna_ids.remove('TGA')

        trna_temp = [x.replace('T', 'U') for x in self.trna_ids]
        self.trna_ids = self.trna_ids + trna_temp




        self.trna_ids_vals = np.linspace(0, 60, 61).astype(int).tolist() +  np.linspace(0, 60, 61).astype(int).tolist()

        self.trna_dict = dict(zip(self.trna_ids, self.trna_ids_vals))
        self.id_to_trna = dict(map(reversed, self.trna_dict.items()))



        # add the U codons
        for key in list(self.human_codon_frequency_bias_nakamura.keys()):
            if 'T' in key:
                val = self.human_codon_frequency_bias_nakamura[key]
                newkey = key.replace('T', 'U')
                self.human_codon_frequency_bias_nakamura[newkey] = val



        self.human_codon_frequency_bias_nakamura_fast = {
            'GCT': 27.7, 'GCC': 27.7, 'GCA': 27.7, 'GCG': 27.7,  #A
            'CGT': 12.2, 'CGC': 12.2, 'CGA': 12.2, 'CGG': 12.2,
            'AGA': 12.2, 'AGG': 12.2,   # R
            'AAT': 19.1, 'AAC': 19.1,   #N
            'GAT': 25.1, 'GAC': 25.1,   # D
            'TGT': 12.6, 'TGC': 12.6,  # C
            'CAA': 34.2, 'CAG': 34.2,  # Q
            'GAA': 39.6, 'GAG': 39.6,  #E
            'GGT': 22.2, 'GGC': 22.2, 'GGA': 22.2, 'GGG': 22.2,  # G
            'CAT': 15.1, 'CAC': 15.1,  # H
            'ATT': 20.8, 'ATC': 20.8, 'ATA': 20.8,  # I
            'TTA': 39.6, 'TTG': 39.6, 'CTT': 39.6, 'CTC': 39.6,
            'CTA': 39.6, 'CTG': 39.6, # L
            'AAA': 31.9, 'AAG': 31.9,  # K
            'ATG': 22.0,   #M
            'TTT': 20.3, 'TTC': 20.3,    # F
            'CCT': 19.8, 'CCC': 19.8, 'CCA': 19.8, 'CCG': 19.8,  # P
            'TCT': 19.5, 'TCC': 19.5, 'TCA': 19.5, 'TCG': 19.5,
            'AGT': 19.5, 'AGC': 19.5,  # S
            'ACT': 18.9, 'ACC': 18.9, 'ACA': 18.9, 'ACG': 18.9, # T
            'TGG': 13.2,   #W
            'TAT': 15.3, 'TAC': 15.3,  # Y
            'GTT': 28.1, 'GTC': 28.1, 'GTA':28.1, 'GTG': 28.1,  # V
            'TAA': 1.6, 'TAG': 1.6, 'TGA':1.6 #STOP
            }


        for key in list(self.human_codon_frequency_bias_nakamura_fast.keys()):
            if 'T' in key:
                val = self.human_codon_frequency_bias_nakamura_fast[key]
                newkey = key.replace('T', 'U')
                self.human_codon_frequency_bias_nakamura_fast[newkey] = val

        self.human_codon_frequency_bias_nakamura_slow = {
            'GCT': 7.4, 'GCC': 7.4, 'GCA': 7.4, 'GCG': 7.4,  #A
            'CGT': 4.5, 'CGC': 4.5, 'CGA': 4.5, 'CGG': 4.5,
            'AGA':4.5, 'AGG':4.5,   #R
            'AAT': 17.0, 'AAC':17.0,  #%N
            'GAT': 21.8, 'GAC': 21.8,  #D
            'TGT': 10.6, 'TGC':10.6,  #C
            'CAA': 12.3, 'CAG': 12.3,  #Q
            'GAA': 29.0, 'GAG': 29.0,  #E
            'GGT': 10.8, 'GGC': 10.8, 'GGA': 10.8, 'GGG': 10.8,  #G
            'CAT': 10.9, 'CAC':10.9,  #H
            'ATT': 7.5, 'ATC': 7.5, 'ATA': 7.5, #I
            'TTA': 7.2, 'TTG':7.2, 'CTT': 7.2, 'CTC': 7.2,
            'CTA': 7.2, 'CTG': 7.2, #L
            'AAA': 24.4, 'AAG': 24.4, #K
            'ATG': 22.0, #M
            'TTT': 17.6, 'TTC': 17.6, #F
            'CCT': 6.9, 'CCC': 6.9, 'CCA': 6.9, 'CCG': 6.9, #P
            'TCT': 4.4, 'TCC': 4.4, 'TCA': 4.4, 'TCG': 4.4,
            'AGT': 4.4, 'AGC': 4.4, #S
            'ACT': 6.1, 'ACC': 6.1, 'ACA': 6.1, 'ACG': 6.1,#T
            'TGG': 13.2, #W
            'TAT': 12.2, 'TAC': 12.2, #Y
            'GTT': 7.1, 'GTC':7.1, 'GTA': 7.1, 'GTG': 7.1, # V
            'TAA': 0.8, 'TAG': 0.8, 'TGA': 0.8 #STOP CODON}
            }

        for key in list(self.human_codon_frequency_bias_nakamura_slow.keys()):
            if 'T' in key:
                val = self.human_codon_frequency_bias_nakamura_slow[key]
                newkey = key.replace('T', 'U')
                self.human_codon_frequency_bias_nakamura_slow[newkey] = val


        self.fast_codons_value = [27.7, 12.2, 19.1, 25.1, 12.6, 34.2, 39.6, 22.2, 15.1,
                                  20.8, 39.6, 31.9, 22, 20.3, 19.8, 19.5,
                                  18.9, 13.2, 15.3, 28.1, 1.6]


        self.slow_codons_value = [7.4, 4.5, 17, 21.8, 10.6, 12.3, 29, 10.8, 10.9, 7.5, 7.2,
                                  24.4, 22, 17.6, 6.9, 4.4, 6.1, 13.2, 12.2, 7.1, .8]

        self.fullcodonkeys = [
            'GCT', 'CGT', 'AAT', 'GAT', 'TGT', 'CAA', 'GAA', 'GGT', 'CAT',
            'ATT', 'TTA', 'AAA', 'ATG', 'TTT', 'CCT', 'TCT',
            'ACT', 'TGG', 'TAT', 'GTT', 'TAA',
            'GCU', 'CGU', 'AAU', 'GAU', 'UGU', 'CAA', 'GAA', 'GGU', 'CAU',
            'AUU', 'UUA', 'AAA', 'AUG', 'UUU', 'CCU', 'TCU',
            'ACU', 'UGG', 'UAU', 'GUU', 'UAA']

        codonkeys = [
            'GCT', 'CGT', 'AAT', 'GAT', 'TGT', 'CAA', 'GAA', 'GGT', 'CAT',
            'ATT', 'TTA', 'AAA', 'ATG', 'TTT', 'CCT', 'TCT',
            'ACT', 'TGG', 'TAT', 'GTT', 'TAA']

        self.sensitivity_fast_slow = []
        for i in range(len(codonkeys)):
            self.sensitivity_fast_slow.append(self.human_codon_frequency_bias_nakamura_fast[codonkeys[i]] / self.human_codon_frequency_bias_nakamura_slow[codonkeys[i]])

        self.load_tags()


        self.strCodonFreq = self.human_codon_frequency_bias_nakamura
        self.strCodonFreq_single = self.human_codon_frequency_bias_nakamura_single
        self.strCodonFreq_fast = self.human_codon_frequency_bias_nakamura_fast
        self.strCodonFreq_slow = self.human_codon_frequency_bias_nakamura_slow

    @property
    def mean_genecopynumber(self):
        '''
        Generate the mean copy number from the human_codon_frequency_bias_nakamura dictionary

        Returns
        -------
        float
            Mean gene copy number.

        '''
        keys = ['TTT', 'TCT', 'TAT', 'TGT', 'TTC', 'TCC', 'TAC', 'TGC', 'TTA', 'TCA', 'TAA', 'TGA', 'TTG', 'TCG', 'TAG', 'TGG', 'CTT', 'CCT', 'CAT', 'CGT', 'CTC', 'CCC', 'CAC', 'CGC', 'CTA', 'CCA', 'CAA', 'CGA', 'CTG', 'CCG', 'CAG', 'CGG', 'ATT', 'ACT', 'AAT', 'AGT', 'ATC', 'ACC', 'AAC', 'AGC', 'ATA', 'ACA', 'AAA', 'AGA', 'ATG', 'ACG', 'AAG', 'AGG', 'GTT', 'GCT', 'GAT', 'GGT', 'GTC', 'GCC', 'GAC', 'GGC', 'GTA', 'GCA', 'GAA', 'GGA', 'GTG', 'GCG', 'GAG', 'GGG']

        return sum([self.human_codon_frequency_bias_nakamura[x] for x in keys ])/len(keys)

    def load_tags(self):
        '''
        Opens custom_tags.txt and adds them to the dictionary

        Returns
        -------
        None.

        '''
        try:
            fname = open("custom_tags.txt", "r")
        except:
            return

        raw = fname.readlines()
        previous_tags = []

        for line in raw:
            if line != '\n':
                previous_tags.append(line)

        for line in previous_tags:

            custom_tag = line.strip('\n').split('---')

            if custom_tag[0] not in self.tag_dict.keys():
                self.tag_dict[custom_tag[0]] = custom_tag[2]

                self.tag_full[custom_tag[0]] = custom_tag[1]
        fname.close()


    def add_custom_tag(self, nt_seq, name):
        '''
        add a custom tag to custom_tags.txt and the tag dictionaries

        Parameters
        ----------
        nt_seq : str
            nucleotide sequence of the tag / epitope.
        name : str
            name of the tag / epitope .

        Returns
        -------
        None.

        '''

        fname = open("custom_tags.txt", "r")

        raw = fname.readlines()
        previous_tags = []

        for line in raw:
            if line != '\n':
                previous_tags.append(line)

        if not set(nt_seq.lower()).issubset(set(['a', 't', 'c', 'g', 'u'])):
            print('invalid NT sequence')
            fname.close()
            return


        aa_seq = ''
        for i in range(0, len(nt_seq), 3):
            aa_seq += self.aa_table[nt_seq[i:i+3]]


        newtag = name+'---'+ nt_seq.lower() + '---'+ aa_seq.upper() + '\n'

        if newtag not in previous_tags:
            previous_tags.append(newtag)
        fname.close()

        fname = open("custom_tags.txt","w+")

        for item in previous_tags:
            fname.write('%s' % item)

        fname.close()
